from collections.abc import Iterable

from kazu.data import Section, Entity
from kazu.utils.spacy_pipeline import SpacyPipelines, BASIC_PIPELINE_NAME, basic_spacy_pipeline
from spacy.tokens import Doc, Span, Token


class KazuToSpacyObjectMapper:
    """Maps entities and text from a :class:`.Section` to the spaCy data model using
    :func:`.basic_spacy_pipeline`\\.

    .. attention::
       Providing incomplete ``entity_classes`` for your usage (or leaving it blank)
       can lead to errors that might only occur infrequently when processing
       the results, and therefore may be difficult to track down.

       Therefore, users should be careful to set ``entity_classes`` to all the entity
       classes corresponding to attributes that they will access on the spaCy
       `Token <https://spacy.io/api/token>`_\\ s within the
       `Span <https://spacy.io/api/span>`_\\ s of the result of :meth:`.__call__`,
       whether directly or via spaCy `Matcher <https://spacy.io/api/matcher>`_ rules
       that check these custom attributes.


       The specific problem is that if you try to read a spaCy custom attribute that
       doesn't exist, you will get an error like::

          AttributeError: [E046] Can't retrieve unregistered extension attribute 'drug'.
          Did you forget to call the `set_extension` method?

       This class uses the provided ``entity_classes`` to call ``set_extension``. If
       the provided ``entity_classes`` is incomplete - say, missing ``"drug"`` - and
       you then try to access the ``drug`` attribute on a token in the result, you will
       get this error.
    """

    def __init__(
        self,
        entity_classes: Iterable[str] = set(),
        set_attributes_incrementally: bool = False,
    ):
        """

        :param entity_classes: known entity classes that the caller intends to access
            the spaCy extension attribute of with the result of :meth:`~.__call__`. See
            note above about the need to take care here.
        :param set_attributes_incrementally: whether to set a spaCy
            `custom extension attribute <https://spacy.io/usage/processing-pipelines#custom-components-attributes>`_
            for 'new' entity classes in :class:`.Section`\\ passed to
            :meth:`.__call__`. This will result in a more consistent result of
            ``__call__``, where every `Span <https://spacy.io/api/span>`_ in the
            dictionary will have an attribute for the relevant :class:`.Entity`'s entity
            class set to ``True`` for all the tokens in the span. However, it makes
            subtle bugs much more likely, so ``False`` is the default - see the note in
            the class-level docs if you are thinking about turning this on.
        """
        #: A set of entity classes known to this class. These will all have a spaCy
        #: `custom extension attribute <https://spacy.io/usage/processing-pipelines#custom-components-attributes>`_
        #: set. If ``set_attributes_incrementally`` is ``True``, as well as the
        #: ``entity_classes`` passed into the ``__init__``, this will include all
        #: entity classes encountered so far processing :class:`.Section`\ s passed in
        #: to :meth:`~.__call__`.
        self.entity_classes = set(entity_classes)
        self.set_attributes_incrementally = set_attributes_incrementally

        for entity_class in self.entity_classes:
            Token.set_extension(entity_class, default=False, force=True)
        self.spacy_pipelines = SpacyPipelines()
        self.spacy_pipelines.add_from_func(BASIC_PIPELINE_NAME, basic_spacy_pipeline)

    def __call__(self, section: Section) -> dict[Entity, Span]:
        """Convert a :class:`.Section` into a dictionary of :class:`.Entity` to
        spaCy `Span <https://spacy.io/api/span>`_\\ s."""

        spacy_doc: Doc = self.spacy_pipelines.process_single(section.text, BASIC_PIPELINE_NAME)
        ent_to_span = {}
        for entity in section.entities:
            entity_class = entity.entity_class
            span = spacy_doc.char_span(
                start_idx=entity.start,
                end_idx=entity.end,
                label=entity_class,
                alignment_mode="expand",
            )
            if span is not None:
                ent_to_span[entity] = span
                if entity_class not in self.entity_classes:
                    if self.set_attributes_incrementally:
                        Token.set_extension(entity_class, default=False, force=True)
                        self.entity_classes.add(entity_class)
                    else:
                        # we don't set the extension, or set the value of it
                        # on the individual tokens.
                        continue
                for token in span:
                    token._.set(entity_class, True)
        return ent_to_span



import logging
from abc import ABC, abstractmethod
from typing import cast, Optional

import pandas as pd
from kazu.data import (
    EquivalentIdSet,
    EquivalentIdAggregationStrategy,
    LinkingCandidate,
    SimpleValue,
    OntologyStringResource,
    AssociatedIdSets,
    GlobalParserActions,
    IdsAndSource,
)
from kazu.database.in_memory_db import (
    MetadataDatabase,
    SynonymDatabase,
)
from kazu.language.string_similarity_scorers import StringSimilarityScorer
from kazu.ontology_preprocessing.autocuration import AutoCurator
from kazu.ontology_preprocessing.curation_utils import (
    OntologyResourceProcessor,
    OntologyStringConflictAnalyser,
    load_ontology_string_resources,
    dump_ontology_string_resources,
    CurationError,
    AutofixStrategy,
    OntologyResourceSetConflictReport,
    OntologyResourceSetMergeReport,
    OntologyResourceSetCompleteReport,
)
from kazu.ontology_preprocessing.ontology_upgrade_report import OntologyUpgradeReport
from kazu.ontology_preprocessing.synonym_generation import CombinatorialSynonymGenerator
from kazu.utils.caching import kazu_disk_cache
from kazu.utils.string_normalizer import StringNormalizer
from kazu.utils.utils import (
    as_path,
    linking_candidates_to_ontology_string_resources,
    PathLike,
)

logger = logging.getLogger(__name__)


#: The column name in a dataframe parsed with :meth:`~.OntologyParser.parse_to_dataframe`
#: for the column of the entity's default/preferred label
DEFAULT_LABEL = "default_label"
#: The column name for the id of each entity
IDX = "idx"
#: The column name for the synonyms/alternative labels for each entity
SYN = "syn"
#: The column name for the type of mapping from default label to synonym - e.g. xref, exactSyn etc. Usually defined by the ontology
MAPPING_TYPE = "mapping_type"
#: The origin of a dataset - e.g. HGNC release 2.1, MEDDRA 24.1 etc.
#: Note, this is different from the parser.name, as is used to identify the origin of a mapping back to a data source
DATA_ORIGIN = "data_origin"

_ONTOLOGY_DEFAULTS_FILENAME = "_defaults.jsonl"
OntologyMetadata = dict[str, dict[str, SimpleValue]]


class OntologyParser(ABC):
    """Parse an ontology (or similar) into a set of outputs suitable for NLP entity
    linking.

    Implementations should have a class attribute 'name' to
    something suitably representative. The key method is
    :meth:`~.OntologyParser.parse_to_dataframe`, which should convert an
    input source to a dataframe suitable for further processing.

    The other important method is :meth:`~.find_kb`. This should parse an ID
    string (if required) and return the underlying source. This is
    important for composite resources that contain identifiers from
    different seed sources.

    See :ref:`ontology_parser` for a more detailed guide.

    Generally speaking, when parsing a data source, synonyms that are
    symbolic (as determined by the :class:`~.StringNormalizer`) that refer to more
    than one id are more likely to be ambiguous. Therefore, we assume
    they refer to unique concepts (e.g. COX 1 could be 'ENSG00000095303'
    OR 'ENSG00000198804', and thus they will yield multiple instances of
    :class:`~.EquivalentIdSet`. Non symbolic synonyms (i.e. noun phrases) are far
    less likely to refer to distinct entities, so we might want to merge
    the associated ID's non-symbolic ambiguous synonyms into a single
    :class:`~.EquivalentIdSet`. The result of :meth:`.StringNormalizer.classify_symbolic`
    forms the ``is_symbolic`` parameter to :meth:`~.score_and_group_ids`.

    If the underlying knowledgebase contains more than one entity type,
    muliple parsers should be implemented, subsetting accordingly (e.g.
    MEDDRA_DISEASE, MEDDRA_DIAGNOSTIC).
    """

    # the synonym table should have these (and only these columns)
    all_synonym_column_names = [IDX, SYN, MAPPING_TYPE]
    # the metadata table should have at least these columns (note, IDX will become the index)
    minimum_metadata_column_names = [DEFAULT_LABEL, DATA_ORIGIN]

    def __init__(
        self,
        in_path: PathLike,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        autocurator: Optional[AutoCurator] = None,
        curations_path: Optional[PathLike] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        """

        :param in_path: Path to some resource that should be processed (e.g. owl file, db config, tsv etc)
        :param entity_class: The entity class to associate with this parser throughout the pipeline.
            Also used in the parser when calling StringNormalizer to determine the class-appropriate behaviour.
        :param name: A string to represent a parser in the overall pipeline. Should be globally unique
        :param string_scorer: Optional protocol of StringSimilarityScorer.  Used for resolving ambiguous symbolic
            synonyms via similarity calculation of the default label associated with the conflicted labels. If no
            instance is provided, all synonym conflicts will be assumed to refer to different concepts. This is not
            recommended!
        :param synonym_merge_threshold: similarity threshold to trigger a merge of conflicted synonyms into a single
            :class:`~.EquivalentIdSet`. See :meth:`~.OntologyParser.score_and_group_ids` for further details
        :param data_origin: The origin of this dataset - e.g. HGNC release 2.1, MEDDRA 24.1 etc. Note, this is different
            from the parser.name, as is used to identify the origin of a mapping back to a data source
        :param synonym_generator: optional CombinatorialSynonymGenerator. Used to generate synonyms for dictionary
            based NER matching
        :param autocurator: optional :class:`~.AutoCurator`. An AutoCurator contains a series of heuristics that
            determines what the default behaviour for a :class:`~.LinkingCandidate` should be. For example, "Ignore
            any strings shorter than two characters or longer than 50 characters", or "use case sensitive matching when
            the LinkingCandidate is symbolic"
        :param curations_path: path to jsonl file of human-curated :class:`~.OntologyStringResource`\\s to override the defaults of the parser.
        :param global_actions: path to json file of :class:`~.GlobalParserActions` to apply to the parser.
        """

        self.in_path = as_path(in_path)
        self.entity_class = entity_class
        self.name = name
        self.ontology_auto_generated_resources_set_path = self.in_path.parent.joinpath(
            f"{self.name}{_ONTOLOGY_DEFAULTS_FILENAME}"
        )
        if string_scorer is None:
            logger.warning(
                "no string scorer configured for %s. Synonym resolution disabled.", self.name
            )
        self.string_scorer = string_scorer
        self.synonym_merge_threshold = synonym_merge_threshold
        self.data_origin = data_origin
        self.synonym_generator = synonym_generator
        self.autocurator = autocurator
        self.curations_path = as_path(curations_path) if curations_path is not None else None
        self.global_actions = global_actions
        self.parsed_dataframe: Optional[pd.DataFrame] = None
        self.metadata_db = MetadataDatabase()
        self.synonym_db = SynonymDatabase()

    @abstractmethod
    def find_kb(self, string: str) -> str:
        """Split an IDX somehow to find the ontology SOURCE reference.

        :param string: the IDX string to process
        :return:
        """
        pass

    def _resolve_candidates(self, candidates_df: pd.DataFrame) -> set[LinkingCandidate]:

        result = set()
        candidates_df["syn_norm"] = candidates_df[SYN].apply(
            StringNormalizer.normalize, entity_class=self.entity_class
        )

        for i, row in (
            candidates_df[["syn_norm", SYN, IDX, MAPPING_TYPE]]
            .groupby(["syn_norm"])
            .agg(set)
            .reset_index()
            .iterrows()
        ):

            syn_set = row[SYN]
            mapping_type_set: frozenset[str] = frozenset(row[MAPPING_TYPE])
            syn_norm = row["syn_norm"]
            if len(syn_set) > 1:
                logger.debug(
                    "normaliser has merged %s into a single candidate: %s", syn_set, syn_norm
                )

            is_symbolic = all(
                StringNormalizer.classify_symbolic(x, self.entity_class) for x in syn_set
            )

            ids: set[str] = row[IDX]
            ids_and_source = set(
                (
                    idx,
                    self.find_kb(idx),
                )
                for idx in ids
            )
            associated_id_sets, agg_strategy = self.score_and_group_ids(ids_and_source, is_symbolic)

            result.add(
                LinkingCandidate(
                    synonym_norm=syn_norm,
                    raw_synonyms=frozenset(syn_set),
                    is_symbolic=is_symbolic,
                    mapping_types=mapping_type_set,
                    associated_id_sets=associated_id_sets,
                    parser_name=self.name,
                    aggregated_by=agg_strategy,
                )
            )

        return result

    def score_and_group_ids(
        self,
        ids_and_source: IdsAndSource,
        is_symbolic: bool,
    ) -> tuple[AssociatedIdSets, EquivalentIdAggregationStrategy]:
        """For a given data source, one normalised synonym may map to one or more id. In
        some cases, the ID may be duplicate/redundant (e.g. there are many chembl ids
        for paracetamol). In other cases, the ID may refer to distinct concepts (e.g.
        COX 1 could be 'ENSG00000095303' OR 'ENSG00000198804').

        Since synonyms from data sources are confused in such a manner, we need to decide some way to cluster them into
        a single :class:`~.LinkingCandidate` concept, which in turn is a container for one or more :class:`~.EquivalentIdSet`
        (depending on whether the concept is ambiguous or not)

        The job of ``score_and_group_ids`` is to determine how many :class:`~.EquivalentIdSet`\\ s for a given set of
        ids should be produced.

        The default algorithm (which can be overridden by concrete parser implementations) works as follows:

        1. If no ``string_scorer`` is configured, create an :class:`~.EquivalentIdSet` for each id (strategy
           :attr:`~.EquivalentIdAggregationStrategy.NO_STRATEGY` - not recommended)
        2. If only one ID is referenced, or the associated normalised synonym string is not symbolic, group the
           ids into a single :class:`~.EquivalentIdSet` (strategy :attr:`~.EquivalentIdAggregationStrategy.UNAMBIGUOUS`)
        3. otherwise, compare the default label associated with each ID to every other default label. If it's above
           ``self.synonym_merge_threshold``, merge into one :class:`~.EquivalentIdSet`, if not, create a new one.

        recommendation: Use the :class:`~.SapbertStringSimilarityScorer` for comparison.

        .. important::
            Any calls to this method requires the metadata DB to be populated, as this is the store of
            :data:`~.DEFAULT_LABEL`.

        :param ids_and_source: ids to determine appropriate groupings of, and their associated sources
        :param is_symbolic: is the underlying synonym symbolic?
        :return:
        """
        if self.string_scorer is None:
            # the NO_STRATEGY aggregation strategy assumes all synonyms are ambiguous
            return (
                frozenset(
                    EquivalentIdSet(ids_and_source=frozenset((single_id_and_source,)))
                    for single_id_and_source in ids_and_source
                ),
                EquivalentIdAggregationStrategy.NO_STRATEGY,
            )
        else:

            if len(ids_and_source) == 1:
                return (
                    frozenset((EquivalentIdSet(ids_and_source=frozenset(ids_and_source)),)),
                    EquivalentIdAggregationStrategy.UNAMBIGUOUS,
                )

            if not is_symbolic:
                return (
                    frozenset((EquivalentIdSet(ids_and_source=frozenset(ids_and_source)),)),
                    EquivalentIdAggregationStrategy.MERGED_AS_NON_SYMBOLIC,
                )
            else:
                # use similarity to group ids into EquivalentIdSets

                DefaultLabels = set[str]
                id_list: list[tuple[IdsAndSource, DefaultLabels]] = []
                for id_and_source_tuple in ids_and_source:
                    default_label = cast(
                        str,
                        self.metadata_db.get_by_idx(self.name, id_and_source_tuple[0])[
                            DEFAULT_LABEL
                        ],
                    )
                    most_similar_id_set = None
                    best_score = 0.0
                    for id_and_default_label_set in id_list:
                        sim = max(
                            self.string_scorer(default_label, other_label)
                            for other_label in id_and_default_label_set[1]
                        )
                        if sim > self.synonym_merge_threshold and sim > best_score:
                            most_similar_id_set = id_and_default_label_set
                            best_score = sim

                    # for the first label, the above for loop is a no-op as id_sets is empty
                    # and the below if statement will be true.
                    # After that, it will be True if the id under consideration should not
                    # merge with any existing group and should get its own EquivalentIdSet
                    if not most_similar_id_set:
                        id_list.append(
                            (
                                {id_and_source_tuple},
                                {default_label},
                            )
                        )
                    else:
                        most_similar_id_set[0].add(id_and_source_tuple)
                        most_similar_id_set[1].add(default_label)

                return (
                    frozenset(
                        EquivalentIdSet(ids_and_source=frozenset(ids_and_source))
                        for ids_and_source, _ in id_list
                    ),
                    EquivalentIdAggregationStrategy.RESOLVED_BY_SIMILARITY,
                )

    def _parse_df_if_not_already_parsed(self):
        if self.parsed_dataframe is None:
            self.parsed_dataframe = self.parse_to_dataframe()
            self.parsed_dataframe[DATA_ORIGIN] = self.data_origin
            self.parsed_dataframe[IDX] = self.parsed_dataframe[IDX].astype(str)
            # since we always need a value for DEFAULT_LABEL,
            # if the underlying data doesn't provide one, just use the IDX
            rows_without_default_label = self.parsed_dataframe.loc[
                pd.isnull(self.parsed_dataframe[DEFAULT_LABEL])
            ]
            rows_without_default_label[DEFAULT_LABEL] = rows_without_default_label[IDX]
            columns_that_should_never_be_empty = [
                IDX
            ] + OntologyParser.minimum_metadata_column_names
            missing_expected_values_mask = (
                self.parsed_dataframe[columns_that_should_never_be_empty].isna().any(axis="columns")
            )

            if missing_expected_values_mask.any():
                raise ValueError(
                    "The parsed dataframe for %s contains missing values expected for metadata. The relevant rows (i.e. entities) will be dropped and ignored: \n\n%s"
                    % self.name,
                    self.parsed_dataframe[missing_expected_values_mask].to_string(
                        max_rows=40, max_cols=10
                    ),
                )

            null_values_per_column = self.parsed_dataframe.isna().any(axis="index")
            columns_with_null_values = null_values_per_column[null_values_per_column].index
            if len(columns_with_null_values) > 0:
                logger.warning(
                    "Some metadata columns have empty values for %s. This is permitted, but if you write custom code to read 'extra' metadata columns"
                    " and depend on the property always being populated, you will likely get errors. Either ensure you can handle this field being 'None',"
                    " or clean the underlying ontology to ensure the value is always populated. Affected columns: \n\n%s",
                    self.name,
                    columns_with_null_values,
                )

    @kazu_disk_cache.memoize(ignore={0})
    def _export_metadata(self, parser_name: str) -> OntologyMetadata:
        """Export the metadata from the ontology.

        :param parser_name: name of this parser. Required for correct operation of cache
            (Note, we cannot pass self to the disk cache as the constructor consumes too
            much memory)
        :return: {idx:{metadata_key:metadata_value}}
        """
        self._parse_df_if_not_already_parsed()
        assert self.parsed_dataframe is not None
        metadata_columns = self.parsed_dataframe.columns
        metadata_columns = metadata_columns.drop([MAPPING_TYPE, SYN])
        metadata_df = self.parsed_dataframe[metadata_columns]
        metadata_df = metadata_df.drop_duplicates(subset=[IDX])
        metadata_df.set_index(inplace=True, drop=True, keys=IDX)
        assert set(OntologyParser.minimum_metadata_column_names).issubset(metadata_df.columns)
        metadata = metadata_df.to_dict(orient="index")
        return cast(OntologyMetadata, metadata)

    def _process_candidates_and_string_resources(
        self, candidates: set[LinkingCandidate], clean_resources: set[OntologyStringResource]
    ) -> tuple[Optional[list[OntologyStringResource]], set[LinkingCandidate]]:

        curation_processor = OntologyResourceProcessor(
            global_actions=self.global_actions,
            resources=list(clean_resources),
            parser_name=self.name,
            entity_class=self.entity_class,
            linking_candidates=candidates,
        )
        return curation_processor.export_resources_and_final_candidates()

    def _resolve_autogenerated_resources(
        self, candidates: set[LinkingCandidate]
    ) -> set[OntologyStringResource]:
        if not self.ontology_auto_generated_resources_set_path.exists():
            clean_resources = self._generate_clean_default_resources(candidates)

        else:
            clean_resources = load_ontology_string_resources(
                self.ontology_auto_generated_resources_set_path
            )
        return clean_resources

    def _create_human_conflict_report(self) -> OntologyResourceSetConflictReport:
        """Creates a report of resource conflicts originating from the human set."""

        assert self.curations_path is not None
        if self.curations_path.exists():
            logger.info(
                "%s curations file found",
                self.name,
            )
            human_curation_set = load_ontology_string_resources(self.curations_path)
        else:
            raise RuntimeError(f"curations not found for {self.name} at {self.curations_path}")

        # set autofix to None so that issues are reported
        conflict_analyser = OntologyStringConflictAnalyser(
            self.entity_class, autofix=AutofixStrategy.NONE
        )

        human_curation_report = conflict_analyser.verify_resource_set_integrity(human_curation_set)

        if len(human_curation_report.normalisation_conflicts) > 0:
            raise CurationError(
                f"{self.name} normalisation conflicts detected in human curation set. Fix these before continuing.\n\n{human_curation_report.normalisation_conflicts}"
            )

        if len(human_curation_report.case_conflicts) > 0:
            logger.warning(
                "%s case conflicts detected in human curation set for %s. These will not be used until they are fixed. See the global conflict report in the model pack root.",
                len(human_curation_report.case_conflicts),
                self.name,
            )
        return human_curation_report

    def _create_merge_report(
        self,
        human_curation_set: set[OntologyStringResource],
        auto_generated_resources_clean: set[OntologyStringResource],
    ) -> OntologyResourceSetMergeReport:
        """Creates a report of how a human resource set will affect an autogenerated
        resource set."""

        # set autofix to None so that issues are reported
        conflict_analyser = OntologyStringConflictAnalyser(
            self.entity_class, autofix=AutofixStrategy.NONE
        )

        return conflict_analyser.merge_human_and_auto_resources(
            human_curated_resources=human_curation_set,
            autocurated_resources=auto_generated_resources_clean,
        )

    def _create_combined_conflict_report(
        self, combined_resource_set: set[OntologyStringResource]
    ) -> OntologyResourceSetConflictReport:
        """Creates a report of resource conflicts originating from the combined human
        and autogenerated set."""
        # set autofix to None so that issues are reported
        conflict_analyser = OntologyStringConflictAnalyser(
            self.entity_class, autofix=AutofixStrategy.NONE
        )
        return conflict_analyser.verify_resource_set_integrity(combined_resource_set)

    def upgrade_ontology_version(self) -> OntologyUpgradeReport:
        """Use when upgrading the version of the underlying ontology data, or when
        changing the configuration of the :class:`.AutoCurator`\\.

        Generate a report that describes the differences in generated :class:`~.OntologyStringResource`\\s
        between the two versions/configurations. Note that this depends on the existence
        of a set of autogenerated :class:`~.OntologyStringResource`\\s from the previous
        configuration.

        To use this method, simply replace the file/directory of the original ontology
        version with the new ontology version in the model pack.

        Note that calling this method will invalidate the disk cache.

        :return:
        """
        if not self.ontology_auto_generated_resources_set_path.exists():
            raise RuntimeError(
                f"{self.name} previous version autogenerated resources not found when asked to build upgrade report, so comparison is not possible.",
            )

        logger.info(
            "loading auto generated resources for previous version of %s",
            self.name,
        )
        original_autogenerated_resources = load_ontology_string_resources(
            self.ontology_auto_generated_resources_set_path
        )
        logger.info(
            "generating resources for new version of %s",
            self.name,
        )
        # need to invalidate cache here, otherwise the previous version of the linking candidates will be returned
        self.clear_cache()
        (
            intermediate_linking_candidates,
            _,
        ) = self._export_metadata_and_intermediate_linking_candidates()
        new_version_auto_generated_resources_clean = self._generate_clean_default_resources(
            intermediate_linking_candidates, save_resources=False
        )
        report = OntologyUpgradeReport(
            previous_version_auto_generated_resources_clean=original_autogenerated_resources,
            new_version_auto_generated_resources_clean=new_version_auto_generated_resources_clean,
        )
        # as a precaution, invalidate cache again in case pipeline is reinstantiated with new invalid cache
        self.clear_cache()
        return report

    def _generate_clean_default_resources(
        self, candidates: set[LinkingCandidate], save_resources: bool = True
    ) -> set[OntologyStringResource]:
        """From a set of candidates, produce default resources that do not conflict
        internally, based upon any configured autocuration regime.

        Note, they may still conflict with other parsers in your model pack.

        :param candidates:
        :param save_resources: should the generated resources be persisted within the
            model pack?
        :return:
        """
        logger.info(
            "%s clean default resources build triggered. This may take some time.",
            self.name,
        )
        # autofix is set to OPTIMISTIC, to ensure a clean set of resources
        conflict_analyser = OntologyStringConflictAnalyser(
            self.entity_class, autofix=AutofixStrategy.OPTIMISTIC
        )
        new_version_auto_generated_resources_dirty = self._generate_dirty_default_resources(
            candidates
        )
        new_version_auto_generated_resources_clean = (
            conflict_analyser.verify_resource_set_integrity(
                new_version_auto_generated_resources_dirty
            ).clean_resources
        )

        logger.info(
            "%s updating auto generated resources in model pack: %s",
            self.name,
            self.ontology_auto_generated_resources_set_path,
        )
        if save_resources:
            dump_ontology_string_resources(
                new_version_auto_generated_resources_clean,
                self.ontology_auto_generated_resources_set_path,
                force=True,
            )

        return new_version_auto_generated_resources_clean

    def _generate_dirty_default_resources(
        self, candidates: set[LinkingCandidate]
    ) -> set[OntologyStringResource]:
        """Dirty resources come directly from a set of :class:`.LinkingCandidate`\\, and
        are optionally further modified by synonym generation and autocuration routines.

        They are not guaranteed to be conflict free - hence why they are 'dirty'.
        """
        default_candidate_set = linking_candidates_to_ontology_string_resources(candidates)
        if self.synonym_generator is not None:
            logger.info(
                "%s synonym generation configuration detected",
                self.name,
            )
            default_candidate_set = self.synonym_generator(default_candidate_set)
        if self.autocurator is not None:
            logger.info(
                "%s autocuration configuration detected",
                self.name,
            )
            default_candidate_set = set(self.autocurator(default_candidate_set))

        return default_candidate_set

    @kazu_disk_cache.memoize(ignore={0})
    def _export_linking_candidates(self, parser_name: str) -> set[LinkingCandidate]:
        """Export :class:`.LinkingCandidate` from the parser.

        :param parser_name: name of this parser. Required for correct operation of cache
            (Note, we cannot pass self to the disk cache as the constructor consumes too
            much memory)
        :return:
        """
        self._parse_df_if_not_already_parsed()
        assert self.parsed_dataframe is not None
        # ensure correct order
        syn_df = self.parsed_dataframe[self.all_synonym_column_names].copy()
        syn_df = syn_df.dropna(subset=[SYN])
        syn_df[SYN] = syn_df[SYN].apply(str.strip)
        syn_df.drop_duplicates(subset=self.all_synonym_column_names)
        assert set(OntologyParser.all_synonym_column_names).issubset(syn_df.columns)
        linking_candidates = self._resolve_candidates(candidates_df=syn_df)
        return linking_candidates

    @kazu_disk_cache.memoize(ignore={0})
    def _populate_databases(
        self, parser_name: str
    ) -> tuple[Optional[list[OntologyStringResource]], OntologyMetadata, set[LinkingCandidate],]:
        """Disk cacheable method that populates all databases.

        :param parser_name: name of this parser. Required for correct operation of cache
            (Note, we cannot pass self to the disk cache as the constructor consumes too
            much memory)
        :return:
        """
        logger.info("populating database for %s from source", self.name)
        metadata, report = self.populate_metadata_db_and_resolve_string_resources()
        (
            maybe_ner_resources,
            final_linking_candidates,
        ) = self._process_candidates_and_string_resources(
            report.intermediate_linking_candidates, report.final_conflict_report.clean_resources
        )
        self.parsed_dataframe = None  # clear the reference to save memory

        self.synonym_db.add_parser(self.name, final_linking_candidates)
        return maybe_ner_resources, metadata, final_linking_candidates

    def populate_metadata_db_and_resolve_string_resources(
        self,
    ) -> tuple[OntologyMetadata, OntologyResourceSetCompleteReport]:
        """Loads the metadata DB and resolves any :class:`~.OntologyStringResource`\\s
        associated with this parser."""

        (
            intermediate_linking_candidates,
            metadata,
        ) = self._export_metadata_and_intermediate_linking_candidates()
        auto_generated_resources_clean = self._resolve_autogenerated_resources(
            intermediate_linking_candidates
        )
        if self.curations_path is None:
            logger.warning(
                "%s is configured to use raw ontology synonyms. This may result in noisy NER performance.",
                self.name,
            )
            return metadata, OntologyResourceSetCompleteReport(
                intermediate_linking_candidates=intermediate_linking_candidates,
                final_conflict_report=OntologyResourceSetConflictReport(
                    clean_resources=auto_generated_resources_clean,
                    merged_resources=set(),
                    normalisation_conflicts=set(),
                    case_conflicts=set(),
                ),
            )
        human_resource_report = self._create_human_conflict_report()

        merge_report = self._create_merge_report(
            human_curation_set=human_resource_report.clean_resources,
            auto_generated_resources_clean=auto_generated_resources_clean,
        )

        combined_report = self._create_combined_conflict_report(merge_report.effective_resources)

        return metadata, OntologyResourceSetCompleteReport(
            intermediate_linking_candidates=intermediate_linking_candidates,
            final_conflict_report=combined_report,
            human_conflict_report=human_resource_report,
            merge_report=merge_report,
        )

    def _export_metadata_and_intermediate_linking_candidates(
        self,
    ) -> tuple[set[LinkingCandidate], OntologyMetadata]:
        """Note that the metadata database loading must be handled here, as the call to
        ``self._export_linking_candidates`` may depend on it being loaded.

        :return:
        """
        metadata = self._export_metadata(self.name)
        # metadata db needs to be populated before call to export_linking_candidates
        self.metadata_db.add_parser(self.name, self.entity_class, metadata)
        intermediate_linking_candidates = self._export_linking_candidates(self.name)
        return intermediate_linking_candidates, metadata

    def populate_databases(
        self, force: bool = False, return_resources: bool = False
    ) -> Optional[list[OntologyStringResource]]:
        """Populate the databases with the results of the parser.

        Also calculates the synonym norms associated with any resources (if provided)
        which can then be used for Dictionary based NER

        :param force: do not use the cache for the ontology parser
        :param return_resources: should processed resources be returned?
        :return: resources if required
        """
        if self.name in self.synonym_db.loaded_parsers and not force and not return_resources:
            logger.debug("parser %s already loaded.", self.name)
            return None

        if force:
            self.clear_cache()

        maybe_resources, metadata, final_linking_candidates = self._populate_databases(self.name)

        if self.name not in self.synonym_db.loaded_parsers:
            logger.info("populating database for %s from cache", self.name)
            self.metadata_db.add_parser(self.name, self.entity_class, metadata)
            self.synonym_db.add_parser(self.name, final_linking_candidates)

        return maybe_resources if return_resources else None

    def clear_cache(self) -> None:
        """Clears the disk cache for this parser."""
        cache_key = self._populate_databases.__cache_key__(self, self.name)
        kazu_disk_cache.delete(self._export_metadata.__cache_key__(self, self.name))
        kazu_disk_cache.delete(self._export_linking_candidates.__cache_key__(self, self.name))
        kazu_disk_cache.delete(cache_key)

    @abstractmethod
    def parse_to_dataframe(self) -> pd.DataFrame:
        """Implementations should override this method, returning a 'long, thin' :class:`pandas.DataFrame` of at least the following
        columns:


        [:data:`~.IDX`, :data:`~.DEFAULT_LABEL`, :data:`~.SYN`, :data:`~.MAPPING_TYPE`]

        | :data:`~.IDX`: the ontology id
        | :data:`~.DEFAULT_LABEL`: the preferred label
        | :data:`~.SYN`: a synonym of the concept
        | :data:`~.MAPPING_TYPE`: the type of mapping from default label to synonym - e.g. xref, exactSyn etc. Usually defined by the ontology

        .. note:: It is the responsibility of the implementation of ``parse_to_dataframe`` to add default labels as synonyms.

        Any 'extra' columns will be added to the :class:`~kazu.database.in_memory_db.MetadataDatabase` as metadata fields for the
        given id in the relevant ontology.
        """
        pass



"""This module contains the core aspects of the :doc:`/datamodel`.

See the page linked above for a quick introduction to the key concepts.

.. |metadata_s11n_warn| replace::

   Note that storing objects here that Kazu can't
   convert to and from json will cause problems for (de)serialization.
   See :ref:`deserialize-generic-metadata` for details.
"""

import json
import uuid
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass, field
from enum import Enum, auto, IntEnum
from math import inf
from typing import Any, Optional, Union

import bson
import cattrs.preconf.json
import cattrs.strategies
from bson import json_util
from kazu.utils.string_normalizer import StringNormalizer
from numpy import ndarray, float32, float16

IS_SUBSPAN = "is_subspan"
# BIO schema
ENTITY_START_SYMBOL = "B"
ENTITY_INSIDE_SYMBOL = "I"
ENTITY_OUTSIDE_SYMBOL = "O"

# key for Document Processing Failed
PROCESSING_EXCEPTION = "PROCESSING_EXCEPTION"


NumericMetric = Union[bool, int, float]
SimpleValue = Union[NumericMetric, str]
JsonEncodable = Optional[Union[dict[str, "JsonEncodable"], list["JsonEncodable"], SimpleValue]]
"""Represents a json-encodable object.

Note that because :class:`dict` is invariant, there can be issues with
using types like ``dict[str, str]`` (see further
`here <https://github.com/python/typing/issues/182#issuecomment-1320974824>`_).
"""


class AutoNameEnum(Enum):
    """Subclass to create an Enum where values are the names when using
    :class:`enum.auto`\\ .

    Taken from the `Python Enum Docs <https://docs.python.org/3/howto/enum.html#using-automatic-values>`_.

    This is `licensed under Zero-Clause BSD. <https://docs.python.org/3/license.html#zero-clause-bsd-license-for-code-in-the-python-release-documentation>`_

    .. raw:: html

        <details>
        <summary>Full License</summary>

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.

    .. raw:: html

        </details>
    """

    def _generate_next_value_(name, start, count, last_values):
        return name


class MentionConfidence(IntEnum):
    HIGHLY_LIKELY = 100  # almost certain to be correct
    PROBABLE = 50
    POSSIBLE = 10  # high degree of uncertainty
    IGNORE = 0  # do not use this mention for NER


class StringMatchConfidence(AutoNameEnum):
    HIGHLY_LIKELY = auto()  # almost certain to be correct
    PROBABLE = auto()  # on the balance of probabilities, will be correct
    POSSIBLE = auto()  # high degree of uncertainty


class DisambiguationConfidence(AutoNameEnum):
    HIGHLY_LIKELY = auto()  # almost certain to be correct
    PROBABLE = auto()  # on the balance of probabilities, will be correct
    POSSIBLE = auto()  # high degree of uncertainty
    AMBIGUOUS = auto()  # could not disambiguate


@dataclass(frozen=True)
class CharSpan:
    """A concept similar to a spaCy Span, except is character index based rather than
    token based.

    Example: ``text[char_span.start:char_span.end]`` will precisely cover the target.
    This means that ``text[char_span.end]`` will give the first character **not** in the
    span.
    """

    start: int
    end: int

    def is_completely_overlapped(self, other: "CharSpan") -> bool:
        """True if other completely overlaps this span.

        :param other:
        :return:
        """
        return self.start >= other.start and self.end <= other.end

    def is_partially_overlapped(self, other: "CharSpan") -> bool:
        """True if other partially overlaps this span.

        :param other:
        :return:
        """
        return (other.start <= self.start <= other.end) or (other.start <= self.end <= other.end)

    def __lt__(self, other):
        return self.start < other.start

    def __gt__(self, other):
        return self.end > other.end


class EquivalentIdAggregationStrategy(AutoNameEnum):
    NO_STRATEGY = auto()  # no strategy. should be used for debugging/testing only
    RESOLVED_BY_SIMILARITY = auto()  # synonym linked to ID via similarity to default ID label
    SYNONYM_IS_AMBIGUOUS = auto()  # synonym has no unambiguous meaning
    CUSTOM = auto()  # a place holder for any strategy that
    UNAMBIGUOUS = auto()
    MERGED_AS_NON_SYMBOLIC = auto()  # used when non-symbolic synonyms are merged
    MODIFIED_BY_CURATION = auto()
    RESOLVED_BY_XREF = auto()


# this is frozen below, but for use in typehints elsewhere the mutable version is
# more useful, as a function written for an immutable frozenset would often behave
# correctly when passed a mutable set, and sometimes our functions actually produce
# the mutable version, as all we wish to do is compare with an immutable version,
# and the mutable version is easier to produce when constructed iteratively.
IdsAndSource = set[tuple[str, str]]


@dataclass(frozen=True, eq=True, order=True)
class EquivalentIdSet:
    """A representation of a set of kb ID's that map to the same synonym and mean the
    same thing."""

    # other ID's mapping to this syn, from different KBs
    ids_and_source: frozenset[tuple[str, str]] = field(default_factory=frozenset)

    @property
    def sources(self) -> set[str]:
        return set(x[1] for x in self.ids_and_source)

    @property
    def ids(self) -> set[str]:
        return set(x[0] for x in self.ids_and_source)


@dataclass(frozen=True)
class Mapping:
    """A mapping is a fully mapped and disambiguated kb concept."""

    #: default label from knowledgebase
    default_label: str
    #: the knowledgebase/database/ontology name
    source: str
    #: the origin of this mapping
    parser_name: str
    #: the identifier within the KB
    idx: str
    string_match_strategy: str
    string_match_confidence: StringMatchConfidence
    disambiguation_confidence: Optional[DisambiguationConfidence] = None
    disambiguation_strategy: Optional[str] = None
    #: source parser name if mapping is an XREF
    xref_source_parser_name: Optional[str] = None
    #: | generic metadata
    #: |
    #: | |metadata_s11n_warn|
    metadata: dict = field(default_factory=dict, hash=False)

    @staticmethod
    def from_dict(mapping_dict: dict) -> "Mapping":
        return kazu_json_converter.structure(mapping_dict, Mapping)


AssociatedIdSets = frozenset[EquivalentIdSet]
"""A frozen set of :class:`.EquivalentIdSet`"""


@dataclass(frozen=True, eq=True)
class LinkingCandidate:
    """A LinkingCandidate is a container for a single normalised synonym, and is
    produced by an :class:`~.OntologyParser` implementation.

    It may be composed of multiple synonyms that normalise to the same
    unique string (e.g. "breast cancer" and "Breast Cancer"). The number
    of ``associated_id_sets`` that this synonym maps to is determined by the
    :meth:`~.OntologyParser.score_and_group_ids` method of the associated OntologyParser.
    """

    #: unnormalised synonym strings
    raw_synonyms: frozenset[str]
    #: normalised form
    synonym_norm: str
    #: ontology parser name
    parser_name: str
    #: is the candidate symbolic? Determined by the OntologyParser
    is_symbolic: bool
    associated_id_sets: AssociatedIdSets
    #: aggregation strategy, determined by the ontology parser
    aggregated_by: EquivalentIdAggregationStrategy = field(hash=False, compare=False)
    #: mapping type metadata
    mapping_types: frozenset[str] = field(hash=False, default_factory=frozenset)

    @property
    def is_ambiguous(self) -> bool:
        return len(self.associated_id_sets) > 1

    @staticmethod
    def from_dict(candidate_dict: dict) -> "LinkingCandidate":
        return kazu_json_converter.structure(candidate_dict, LinkingCandidate)


@dataclass()
class LinkingMetrics:
    """Metrics for Entity Linking.

    LinkingMetrics holds data on various quality metrics, for how well a :class:`~.LinkingCandidate`
     maps to a host :class:`~.Entity`\\.
    """

    search_score: Optional[float] = field(compare=False, default=None)
    embed_score: Optional[float] = field(compare=False, default=None)
    bool_score: Optional[bool] = field(compare=False, default=None)
    exact_match: Optional[bool] = field(compare=False, default=None)

    @staticmethod
    def from_dict(metric_dict: dict) -> "LinkingMetrics":
        return kazu_json_converter.structure(metric_dict, LinkingMetrics)


CandidatesToMetrics = dict[LinkingCandidate, LinkingMetrics]
"""This type is used whenever we have :class:`~.LinkingCandidate`\\ s and some metrics
for how well they map to a specific :class:`~.Entity`\\.

In particular, :attr:`~.Entity.linking_candidates` holds relevant candidates and their metrics, and this type is used in parts
of kazu which produce, modify or use this field.
"""


@dataclass(unsafe_hash=True)
class Entity:
    """A :class:`kazu.data.Entity` is a container for information about a
    single entity detected within a :class:`kazu.data.Section`\\ .

    Within an :class:`kazu.data.Entity`, the most important fields are :attr:`.Entity.match` (the actual string detected),
    :attr:`.Entity.linking_candidates`, (candidates for knowledgebase hits)
    and :attr:`.Entity.mappings`, the final product of linked references to the underlying entity.
    """

    #: exact text representation
    match: str
    entity_class: str
    spans: frozenset[CharSpan]
    #: namespace of the :class:`~.Step` that produced this instance
    namespace: str
    mention_confidence: MentionConfidence = MentionConfidence.HIGHLY_LIKELY
    _id: str = field(default_factory=lambda: uuid.uuid4().hex)
    mappings: set[Mapping] = field(default_factory=set, hash=False)
    #: | generic metadata
    #: |
    #: | |metadata_s11n_warn|
    metadata: dict = field(default_factory=dict, hash=False)
    start: int = field(init=False, hash=False)
    end: int = field(init=False, hash=False)
    match_norm: str = field(init=False, hash=False)
    linking_candidates: CandidatesToMetrics = field(default_factory=dict, hash=False)

    def add_or_update_linking_candidates(self, candidates: CandidatesToMetrics) -> None:
        for candidate, metrics in candidates.items():
            self.add_or_update_linking_candidate(candidate, metrics)

    def add_or_update_linking_candidate(
        self, candidate: LinkingCandidate, new_metrics: LinkingMetrics
    ) -> None:
        maybe_existing_metrics = self.linking_candidates.get(candidate)
        if not maybe_existing_metrics:
            self.linking_candidates[candidate] = new_metrics
        else:
            for k, v in new_metrics.__dict__.items():
                if v is not None:
                    setattr(maybe_existing_metrics, k, v)

    def calc_starts_and_ends(self) -> tuple[int, int]:
        earliest_start = inf
        latest_end = 0
        for span in self.spans:
            if span.start < earliest_start:
                earliest_start = span.start
            if span.end > latest_end:
                latest_end = span.end
        if earliest_start is inf:
            raise ValueError("spans are not valid")
        return int(earliest_start), latest_end

    def __post_init__(self):
        self.start, self.end = self.calc_starts_and_ends()
        self.match_norm = StringNormalizer.normalize(self.match, self.entity_class)

    def is_completely_overlapped(self, other: "Entity") -> bool:
        """True if all CharSpan instances are completely encompassed by all other
        CharSpan instances.

        :param other:
        :return:
        """
        for charspan in self.spans:
            for other_charspan in other.spans:
                if charspan.is_completely_overlapped(other_charspan):
                    break
            else:
                return False
        return True

    def is_partially_overlapped(self, other: "Entity") -> bool:
        """True if only one CharSpan instance is defined in both self and other, and
        they are partially overlapped.

        If multiple CharSpan are defined in both self and other, this becomes pathological, as while they may overlap
        in the technical sense, they may have distinct semantic meaning. For instance, consider the case where
        we may want to use is_partially_overlapped to select the longest annotation span suggested
        by some NER system.

        case 1:
        text: the patient has metastatic liver cancers
        entity1:  metastatic liver cancer -> [CharSpan(16,39]
        entity2: liver cancers -> [CharSpan(27,40]

        result: is_partially_overlapped -> True (entities are part of same concept)


        case 2: non-contiguous entities

        text: lung and liver cancer
        lung cancer -> [CharSpan(0,4), CharSpan(15, 21)]
        liver cancer -> [CharSpan(9,21)]

        result: is_partially_overlapped -> False (entities are distinct)

        :param other:
        :return:
        """
        if len(self.spans) == 1 and len(other.spans) == 1:
            (charspan,) = self.spans
            (other_charspan,) = other.spans
            return charspan.is_partially_overlapped(other_charspan)
        else:
            return False

    def __len__(self) -> int:
        """Span length.

        :return: number of characters enclosed by span
        """
        return self.end - self.start

    def __repr__(self) -> str:
        """Describe the tag.

        :return: tag match description
        """
        return f"{self.match}:{self.entity_class}:{self.namespace}:{self.start}:{self.end}"

    def as_brat(self) -> str:
        """
        :return: this entity in the third party biomedical nlp Brat format
            (see the `docs <https://brat.nlplab.org/introduction.html>`_\\ ,
            `paper <https://aclanthology.org/E12-2021.pdf>`_\\ , and
            `codebase <https://github.com/nlplab/brat>`_\\ )
        """
        # TODO: update this to make use of non-contiguous entities
        return f"{hash(self)}\t{self.entity_class}\t{self.start}\t{self.end}\t{self.match}\n"

    def add_mapping(self, mapping: Mapping) -> None:
        """Deprecated.

        :param mapping:
        :return:
        """
        self.mappings.add(mapping)

    @classmethod
    def from_spans(
        cls, spans: list[tuple[int, int]], text: str, join_str: str = "", **kwargs: Any
    ) -> "Entity":
        """Create an instance of Entity from a list of character indices. A text string
        of underlying doc is also required to produce a representative match.

        :param spans:
        :param text:
        :param join_str: a string used to join the spans together
        :param kwargs:
        :return:
        """
        char_spans = []
        text_pieces = []
        for start, end in spans:
            text_pieces.append(text[start:end])
            char_spans.append(CharSpan(start=start, end=end))
        return cls(spans=frozenset(char_spans), match=join_str.join(text_pieces), **kwargs)

    @classmethod
    def load_contiguous_entity(cls, start: int, end: int, **kwargs: Any) -> "Entity":
        single_span = frozenset([CharSpan(start=start, end=end)])
        return cls(spans=single_span, **kwargs)

    @staticmethod
    def from_dict(entity_dict: dict) -> "Entity":
        return kazu_json_converter.structure(entity_dict, Entity)


@dataclass(unsafe_hash=True)
class Section:
    """A container for text and entities.

    One or more make up a :class:`kazu.data.Document`.
    """

    #: the text to be processed
    text: str
    #: the name of the section (e.g. abstract, body, header, footer etc)
    name: str
    #: | generic metadata
    #: |
    #: | |metadata_s11n_warn|
    metadata: dict = field(default_factory=dict, hash=False)
    #: entities detected in this section
    entities: list[Entity] = field(default_factory=list, hash=False)
    _sentence_spans: Optional[tuple[CharSpan, ...]] = field(
        default=None, hash=False, init=False
    )  # hidden implem. to prevent overwriting existing sentence spans

    @property
    def sentence_spans(self) -> Iterable[CharSpan]:
        if self._sentence_spans is not None:
            return self._sentence_spans
        else:
            return ()

    @sentence_spans.setter
    def sentence_spans(self, sent_spans: Iterable[CharSpan]) -> None:
        """Setter for sentence_spans. sentence_spans are stored in the order provided by
        the iterable sent_spans param, which may not necessarily be in sorted order.

        :param sent_spans:
        :return:
        """
        if self._sentence_spans is None:
            sent_spans_tuple = tuple(sent_spans)
            assert len(sent_spans_tuple) == len(
                set(sent_spans_tuple)
            ), "There are duplicate sentence spans"
            self._sentence_spans = sent_spans_tuple
        else:
            raise AttributeError("Immutable sentence_spans is already set")

    def __str__(self):
        return f"name: {self.name}, text: {self.text[:100]}"

    @staticmethod
    def from_dict(section_dict: dict) -> "Section":
        return kazu_json_converter.structure(section_dict, Section)


@dataclass(unsafe_hash=True)
class Document:
    """A container that is the primary input into a :class:`kazu.pipeline.Pipeline`."""

    #: a document identifier
    idx: str = field(default_factory=lambda: uuid.uuid4().hex)
    #: sections comprising this document
    sections: list[Section] = field(default_factory=list, hash=False)
    #: | generic metadata
    #: |
    #: | |metadata_s11n_warn|
    metadata: dict = field(default_factory=dict, hash=False)

    def __str__(self):
        return f"idx: {self.idx}"

    def get_entities(self) -> list[Entity]:
        """Get all entities in this document."""
        entities = []
        for section in self.sections:
            entities.extend(section.entities)
        return entities

    def to_json(
        self,
        **kwargs: Any,
    ) -> str:
        """Convert to json string.

        :param kwargs: passed through to :func:`json.dumps`.
        :return:
        """
        return kazu_json_converter.dumps(self, **kwargs)

    def to_dict(self) -> dict:
        """Convert the Document to a ``dict``."""
        # type ignore needed because cattrs says this could be 'any', but we know more specifically it will be a
        # json-encodable dict
        return kazu_json_converter.unstructure(self)  # type: ignore[no-any-return]

    @classmethod
    def create_simple_document(cls, text: str) -> "Document":
        """Create an instance of :class:`.Document` from a text string.

        :param text:
        :return:
        """
        return cls(sections=[Section(text=text, name="na")])

    @classmethod
    def simple_document_from_sents(cls, sents: list[str]) -> "Document":
        section = Section(text=" ".join(sents), name="na")
        sent_spans = []
        curr_start = 0
        for sent in sents:
            sent_spans.append(CharSpan(start=curr_start, end=curr_start + len(sent)))
            curr_start += len(sent) + 1  # + 1 is for the joining space
        section.sentence_spans = sent_spans
        return cls(sections=[section])

    @classmethod
    def from_named_section_texts(cls, named_sections: dict[str, str]) -> "Document":
        sections = [Section(text=text, name=name) for name, text in named_sections.items()]
        return cls(sections=sections)

    def __len__(self):
        return sum(len(section.text) for section in self.sections)

    @staticmethod
    def from_dict(document_dict: dict) -> "Document":
        return kazu_json_converter.structure(document_dict, Document)

    @staticmethod
    def from_json(json_str: str) -> "Document":
        return Document.from_dict(json.loads(json_str))


def _initialize_json_converter(testing: bool = False) -> cattrs.preconf.json.JsonConverter:
    json_conv = cattrs.preconf.json.make_converter(omit_if_default=True, forbid_extra_keys=True)

    # 'external' to kazu datatypes
    json_conv.register_unstructure_hook(float16, lambda v: v.item())
    json_conv.register_structure_hook(float16, lambda v, _: float16(v))
    json_conv.register_unstructure_hook(float32, lambda v: v.item())
    json_conv.register_structure_hook(float32, lambda v, _: float16(v))
    json_conv.register_unstructure_hook(ndarray, lambda v: v.tolist())
    # note: this could result in a change in the dtype (e.g. float vs int and precision)
    # from the 'original' if roundtripped. This doesn't seem like a deal breaker for
    # what we're using it for.
    json_conv.register_structure_hook(ndarray, lambda v, _: ndarray(v))
    json_conv.register_unstructure_hook(bson.ObjectId, lambda v: json_util.default(v))
    json_conv.register_structure_hook(bson.ObjectId, lambda v, _: json_util.object_hook(v))

    json_conv.register_unstructure_hook(MentionConfidence, lambda v: v.name)
    json_conv.register_structure_hook(MentionConfidence, lambda v, _: MentionConfidence[v])

    def _linking_candidates_unstruct_hook(
        candidates_and_metrics: CandidatesToMetrics,
    ) -> list[list[dict[str, JsonEncodable]]]:
        return [
            [json_conv.unstructure(candidate), json_conv.unstructure(metrics)]
            for candidate, metrics in candidates_and_metrics.items()
        ]

    json_conv.register_unstructure_hook(
        Entity,
        cattrs.gen.make_dict_unstructure_fn(
            Entity,
            json_conv,
            # omit the `_id` if we're not testing, as it's an 'internal' field
            # that users won't want to see in serialized output by default,
            # and it will add to the output size.
            _id=cattrs.gen.override(omit=not testing),
            linking_candidates=cattrs.gen.override(unstruct_hook=_linking_candidates_unstruct_hook),
            _cattrs_include_init_false=True,
            _cattrs_omit_if_default=True,
        ),
    )

    def _linking_candidates_struct_hook(
        candidates_and_metrics: list[list[dict[str, JsonEncodable]]],
        _: type,
    ) -> CandidatesToMetrics:
        return {
            json_conv.structure(candidate_and_metric[0], LinkingCandidate): json_conv.structure(
                candidate_and_metric[1], LinkingMetrics
            )
            for candidate_and_metric in candidates_and_metrics
        }

    json_conv.register_structure_hook(
        Entity,
        cattrs.gen.make_dict_structure_fn(
            Entity,
            json_conv,
            linking_candidates=cattrs.gen.override(struct_hook=_linking_candidates_struct_hook),
            _cattrs_include_init_false=True,
        ),
    )

    json_conv.register_unstructure_hook(
        Section,
        cattrs.gen.make_dict_unstructure_fn(
            Section,
            json_conv,
            _sentence_spans=cattrs.gen.override(rename="sentence_spans", omit_if_default=True),
            _cattrs_include_init_false=True,
            _cattrs_omit_if_default=True,
        ),
    )

    json_conv.register_structure_hook(
        Section,
        cattrs.gen.make_dict_structure_fn(
            Section,
            json_conv,
            _sentence_spans=cattrs.gen.override(
                rename="sentence_spans",
                # needed because cattrs by default ignores init=False fields
                omit=False,
                struct_hook=lambda v, _: (
                    None if len(v) == 0 else tuple(json_conv.structure(x, CharSpan) for x in v)
                ),
            ),
            _cattrs_include_init_false=True,
        ),
    )
    return json_conv


kazu_json_converter = _initialize_json_converter(testing=False)
"""A `cattrs Converter <https://catt.rs/en/stable/converters.html>`_
configured for converting Kazu's datamodel into json.

If you are not familiar with ``cattrs``, don't worry: you can just
use methods on the kazu classes like :meth:`Document.from_dict` and
:meth:`Document.from_dict`, and you will likely never need to use or understand
``kazu_json_converter``.

If you are familiar with cattrs, you may prefer to use the ``structure``, ``unstructure``,
``dumps`` and ``loads`` methods of ``kazu_json_converter`` directly.
"""


class OntologyStringBehaviour(AutoNameEnum):
    #: use the resource for both dictionary based NER and as a linking target.
    ADD_FOR_NER_AND_LINKING = auto()
    #: use the resource only as a linking target. Note, this is not required if the resource is already in the
    #: underlying ontology, as all ontology resources are included as linking targets by default
    #: (Also see DROP_FOR_LINKING)
    ADD_FOR_LINKING_ONLY = auto()
    #: do not use this resource as a linking target. Normally, you would use this for a resource you want to remove
    #: from the underlying ontology (e.g. a 'bad' synonym). If the resource does not exist, has no effect
    DROP_FOR_LINKING = auto()


class ParserBehaviour(AutoNameEnum):
    #: completely remove the ids from the parser - i.e. should never used anywhere
    DROP_IDS_FROM_PARSER = auto()


@dataclass(frozen=True)
class ParserAction:
    """A ParserAction changes the behaviour of a
    :class:`kazu.ontology_preprocessing.base.OntologyParser` in a global sense.

    A ParserAction overrides any default behaviour of the parser, and also any conflicts that may occur with
    :class:`.OntologyStringResource`\\s.

    These actions are useful for eliminating unwanted behaviour. For example, the root of the Mondo
    ontology is http://purl.obolibrary.org/obo/HP_0000001, which has a default label of 'All'. Since this is
    such a common word, and not very useful in terms of linking, we might want a global action so that this
    ID is not used anywhere in a Kazu pipeline.

    The parser_to_target_id_mappings field should specify the parser name and an affected IDs if required.
    See :class:`.ParserBehaviour` for the type of actions that are possible.
    """

    behaviour: ParserBehaviour
    parser_to_target_id_mappings: dict[str, set[str]] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, json_dict: dict) -> "ParserAction":
        return kazu_json_converter.structure(json_dict, ParserAction)

    def __post_init__(self):
        if len(self.parser_to_target_id_mappings) == 0:
            raise ValueError(
                f"parser_to_target_id_mappings must be specified for global actions: {self}"
            )
        for key, values in self.parser_to_target_id_mappings.items():
            if len(values) == 0:
                raise ValueError(f"at least one ID must be specified for key: {key}, {self}")


@dataclass(frozen=True)
class GlobalParserActions:
    """Container for all :class:`.ParserAction`\\s."""

    actions: list[ParserAction]
    _parser_name_to_action: defaultdict[str, list[ParserAction]] = field(
        default_factory=lambda: defaultdict(list), init=False
    )

    def __post_init__(self):
        for action in self.actions:
            for parser_name in action.parser_to_target_id_mappings.keys():
                self._parser_name_to_action[parser_name].append(action)

    def parser_behaviour(self, parser_name: str) -> Iterable[ParserAction]:
        """Generator that yields behaviours for a specific parser, based on the order
        they are specified in.

        :param parser_name:
        :return:
        """
        yield from self._parser_name_to_action.get(parser_name, [])

    @classmethod
    def from_dict(cls, json_dict: dict) -> "GlobalParserActions":
        return kazu_json_converter.structure(json_dict, GlobalParserActions)


@dataclass(frozen=True)
class Synonym:
    text: str
    case_sensitive: bool
    mention_confidence: MentionConfidence


@dataclass(frozen=True)
class OntologyStringResource:
    """A OntologyStringResource represents the behaviour of a specific
    :class:`.LinkingCandidate` within an Ontology.

    For each LinkingCandidate, a default OntologyStringResource is produced with its
    behaviour determined by an instance of
    :class:`kazu.ontology_preprocessing.autocuration.AutoCurator` and the
    :class:`kazu.ontology_preprocessing.curation_utils.OntologyStringConflictAnalyser`\\.

    .. note::

       This is typically handled by the internals of :class:`kazu.ontology_preprocessing.base.OntologyParser`\\.
       However, OntologyStringResources can also be used to override the default behaviour of a parser. See :ref:`ontology_parser`
       for a more detailed guide.

    The configuration of a OntologyStringResource will affect both NER and Linking aspects of Kazu:

    Example 1:

    The string 'ALL' is highly ambiguous. It might mean several diseases, or simply 'all'. Therefore, we want
    to add a curation as follows, so that it will only be used as a linking target and not for dictionary based NER:

    .. code-block:: python

        OntologyStringResource(
            original_synonyms=frozenset(
                [
                    Synonym(
                        text="ALL",
                        mention_confidence=MentionConfidence.POSSIBLE,
                        case_sensitive=True,
                    )
                ]
            ),
            behaviour=OntologyStringBehaviour.ADD_FOR_LINKING_ONLY,
        )


    Example 2:

    The string 'LH' is incorrectly identified as a synonym of the PLOD1 (ENSG00000083444) gene, whereas more often than not, it's actually an abbreviation of Lutenising Hormone.
    We therefore want to override the associated_id_sets to LHB (ENSG00000104826, or Lutenising Hormone Subunit Beta)

    The OntologyStringResource we therefore want is:

    .. code-block:: python

        OntologyStringResource(
            original_synonyms=frozenset(
                [
                    Synonym(
                        text="LH",
                        mention_confidence=MentionConfidence.POSSIBLE,
                        case_sensitive=True,
                    )
                ]
            ),
            associated_id_sets=frozenset((EquivalentIdSet(("ENSG00000104826", "ENSEMBL")),)),
            behaviour=OntologyStringBehaviour.ADD_FOR_LINKING_ONLY,
        )

    Example 3:

    A :class:`.LinkingCandidate` has an alternative synonym not referenced in the underlying ontology, and we want to add it.

    .. code-block:: python

        OntologyStringResource(
            original_synonyms=frozenset(
                [
                    Synonym(
                        text="breast carcinoma",
                        mention_confidence=MentionConfidence.POSSIBLE,
                        case_sensitive=True,
                    )
                ]
            ),
            associated_id_sets=frozenset((EquivalentIdSet(("ENSG00000104826", "ENSEMBL")),)),
            behaviour=OntologyStringBehaviour.ADD_FOR_NER_AND_LINKING,
        )
    """

    #: Original synonyms, exactly as specified in the source ontology. These should all normalise to the same string.
    original_synonyms: frozenset[Synonym]
    #: The intended behaviour for this resource.
    behaviour: OntologyStringBehaviour
    #: Alternative synonyms generated from the originals by :class:`kazu.ontology_preprocessing.synonym_generation.CombinatorialSynonymGenerator`\.
    alternative_synonyms: frozenset[Synonym] = field(default_factory=frozenset)
    #: If specified, will override the parser defaults for the associated :class:`.LinkingCandidate`\, as long as conflicts do not occur
    associated_id_sets: Optional[AssociatedIdSets] = None
    _id: bson.ObjectId = field(default_factory=bson.ObjectId, compare=False)
    #: results of any decisions by the :class:`kazu.ontology_preprocessing.autocuration.AutoCurator`
    autocuration_results: Optional[dict[str, str]] = field(default=None, compare=False)
    #: human readable comments about this curation decision
    comment: Optional[str] = field(default=None, compare=False)

    def __post_init__(self):
        cs_syns = defaultdict(set)
        ci_syns = defaultdict(set)
        if len(self.original_synonyms) == 0:
            raise ValueError(f"no synonyms for: {self}")
        for syn in self.active_ner_synonyms():
            if syn.case_sensitive:
                cs_syns[syn.text].add(syn.mention_confidence)
            else:
                ci_syns[syn.text.lower()].add(syn.mention_confidence)
        for cs_syn, cs_confidences in ci_syns.items():
            ci_confidences = ci_syns.get(cs_syn.lower(), {MentionConfidence.POSSIBLE})
            if min(ci_confidences) < min(cs_confidences):
                raise ValueError(f"case sensitive conflict: {self}")

    def syn_norm_for_linking(self, entity_class: str) -> str:
        norms = set(
            StringNormalizer.normalize(syn.text, entity_class) for syn in self.original_synonyms
        )
        if len(norms) == 1:
            syn_norm = next(iter(norms))
            return syn_norm
        else:
            raise ValueError(
                f"multiple synonym norms produced by {self}. This resource should be separated into two or more separate items."
            )

    @staticmethod
    def from_json(json_str: str) -> "OntologyStringResource":
        json_dict = json.loads(json_str)
        return OntologyStringResource.from_dict(json_dict)

    @classmethod
    def from_dict(cls, json_dict: dict) -> "OntologyStringResource":
        return kazu_json_converter.structure(json_dict, OntologyStringResource)

    def to_dict(self, preserve_structured_object_id: bool = True) -> dict[str, Any]:
        as_dict: dict[str, Any] = kazu_json_converter.unstructure(self)
        if preserve_structured_object_id:
            as_dict["_id"] = self._id
        return as_dict

    def to_json(self) -> str:
        as_json = self.to_dict(False)
        assert isinstance(as_json, dict)
        return json.dumps(as_json)

    @property
    def additional_to_source(self) -> bool:
        """True if this resource created in addition to the source resources defined in
        the original Ontology."""
        return self.associated_id_sets is not None and self.behaviour in {
            OntologyStringBehaviour.ADD_FOR_NER_AND_LINKING,
            OntologyStringBehaviour.ADD_FOR_LINKING_ONLY,
        }

    def all_synonyms(self) -> Iterable[Synonym]:
        for syn in self.original_synonyms.union(self.alternative_synonyms):
            yield syn

    def all_strings(self) -> Iterable[str]:
        for syn in self.all_synonyms():
            yield syn.text

    def active_ner_synonyms(self) -> Iterable[Synonym]:
        if self.behaviour is OntologyStringBehaviour.ADD_FOR_NER_AND_LINKING:
            for syn in self.original_synonyms.union(self.alternative_synonyms):
                if syn.mention_confidence is not MentionConfidence.IGNORE:
                    yield syn


class KazuConfigurationError(Exception):
    pass



from typing import TypedDict

KAZU = "kazu"
API = "api"


class _FastAPIDocsUrls(TypedDict):
    docs_url: str
    openapi_url: str
    redoc_url: str


API_DOCS_URLS: _FastAPIDocsUrls = {
    "docs_url": f"/{API}/docs",  # this is the Swagger page
    "openapi_url": f"/{API}/openapi.json",
    "redoc_url": f"/{API}/redoc",
}

_API_KAZU_PREFIX = f"/{API}/{KAZU}/"

STEPS = f"/{API}/steps"
STEP_GROUPS = f"/{API}/step_groups"
NER_AND_LINKING = f"{_API_KAZU_PREFIX}ner_and_linking"
CUSTOM_PIPELINE_STEPS = f"{_API_KAZU_PREFIX}custom_pipeline_steps"
NER_ONLY = f"{_API_KAZU_PREFIX}ner_only"
LINKING_ONLY = f"{_API_KAZU_PREFIX}linking_only"
BATCH = f"{_API_KAZU_PREFIX}batch"
LS_ANNOTATIONS = f"{_API_KAZU_PREFIX}ls-annotations"
STEP_GROUP_WITH_VAR = f"{_API_KAZU_PREFIX}{{step_group}}"

NO_AUTH_ENDPOINTS = [
    "/",
    f"/{API}",
    f"/{API}/",
    LS_ANNOTATIONS,
    STEPS,
    STEP_GROUPS,
]

# type ignore necessary because the way TypedDict currently works,
# one of the values could be something other than a string, as
# we could be using a subclass of _FastAPIDocsUrls which adds a field
# with some other type. See https://github.com/python/mypy/issues/7981
# for discussion and possible solution
NO_AUTH_ENDPOINTS.extend(API_DOCS_URLS.values())  # type: ignore[arg-type]

NO_AUTH_DIRS = [
    # nothing under /ui requires auth
    # the main page is /ui/index, but
    # it needs to be able to load static resources
    # which are places under /ui as well.
    "/ui"
]

AUTH_ENDPOINTS = [
    NER_AND_LINKING,
    CUSTOM_PIPELINE_STEPS,
    NER_ONLY,
    LINKING_ONLY,
    BATCH,
    STEP_GROUP_WITH_VAR,
    f"/{API}/{KAZU}",  # old legacy single-doc api
]



import functools
import json
import logging
import pickle
from collections.abc import Iterable
from os import getenv
from pathlib import Path
from typing import cast, Optional

import numpy as np
from scipy.sparse import csr_matrix, vstack
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.utils.extmath import safe_sparse_dot

from kazu.data import EquivalentIdSet
from kazu.database.in_memory_db import SynonymDatabase
from kazu.utils.caching import kazu_disk_cache
from kazu.utils.utils import create_char_ngrams, create_word_ngrams, Singleton

logger = logging.getLogger(__name__)


def create_word_and_char_ngrams(
    s: str,
    words: Iterable[int] = (
        1,
        2,
    ),
    chars: Iterable[int] = (
        2,
        3,
    ),
) -> list[str]:
    """Function to create char and word ngrams.

    :param s: string to process
    :param words: create n words
    :param chars: create n chars
    :return: list of strings comprised of words and chars
    """
    strings = []
    for ngram_word in words:
        strings.extend(create_word_ngrams(s, ngram_word))
    for ngram_char in chars:
        strings.extend(create_char_ngrams(s, ngram_char))
    return strings


class TfIdfScorer(metaclass=Singleton):
    """This class manages a set of TFIDF models (via
    :class:`sklearn.feature_extraction.text.TfidfVectorizer`\\ ).

    It's a singleton, so that the models can be accessed in multiple locations without
    the need to load them into memory multiple times.
    """

    def __init__(self):
        self.synonym_db = SynonymDatabase()
        self.parser_to_vectorizer: dict[str, TfidfVectorizer] = self.build_vectorizers()

    @kazu_disk_cache.memoize(ignore={0})
    def build_vectorizers(self) -> dict[str, TfidfVectorizer]:
        result: dict[str, TfidfVectorizer] = {}
        for parser_name in self.synonym_db.loaded_parsers:
            synonyms = self.synonym_db.get_all(parser_name).keys()
            vectoriser = TfidfVectorizer(lowercase=False, analyzer=create_word_and_char_ngrams)
            vectoriser.fit(synonyms)
            result[parser_name] = vectoriser
        return result

    def __call__(
        self, strings: list[str], matrix: np.ndarray, parser: str
    ) -> Iterable[tuple[str, float]]:
        """Transform a list of strings with a parser-specific vectorizer and score
        against a matrix.

        :param strings:
        :param matrix:
        :param parser:
        :return: matching strings and their score sorted by best score
        """
        if len(strings) == 1:
            yield strings[0], 100.0
        else:
            mat = self.parser_to_vectorizer[parser].transform(strings)
            score_matrix = np.squeeze(-np.asarray(mat.dot(matrix.T).todense()))
            neighbours = score_matrix.argsort()
            for neighbour in neighbours:
                yield strings[neighbour], -score_matrix[neighbour]


class GildaTfIdfScorer(metaclass=Singleton):
    """This class uses a single TFIDF model for 'Gilda-inspired' method of
    disambiguation. It uses a pretrained TF-IDF model, and contexual text mapped to
    knowledgebase identifiers (such as wikipedia descriptions of the entity). The sparse
    matrices of these contexts are then compared cosine wise with a target matrix to
    determine the most likely identifier.

    Context matrices are kept in a disk cache until needed, with only a sample held in memory.
    The size of this in memory cache can be controlled with the
    ``KAZU_GILDA_TFIDF_DISAMBIGUATION_IN_MEMORY_CACHE_SIZE`` env variable.

    .. caution::
       If no context is available, the ID automatically scores 0.0.
       The downside of this is that any ids without a context automatically
       appear at the bottom of any rankings.

    Original Credit:

    https://github.com/indralab/gilda

    Paper:

    | Benjamin M Gyori, Charles Tapley Hoyt, and Albert Steppi. 2022.
    | `Gilda: biomedical entity text normalization with machine-learned disambiguation as a service. <https://doi.org/10.1093/bioadv/vbac034>`_
    | Bioinformatics Advances. Vbac034.

    .. raw:: html

        <details>
        <summary>Bibtex Citation Details</summary>

    .. code:: bibtex

        @article{gyori2022gilda,
            author = {Gyori, Benjamin M and Hoyt, Charles Tapley and Steppi, Albert},
            title = "{{Gilda: biomedical entity text normalization with machine-learned disambiguation as a service}}",
            journal = {Bioinformatics Advances},
            year = {2022},
            month = {05},
            issn = {2635-0041},
            doi = {10.1093/bioadv/vbac034},
            url = {https://doi.org/10.1093/bioadv/vbac034},
            note = {vbac034}
        }

    .. raw:: html

        </details>

    It's a singleton, so that the model can be accessed in multiple locations without the need to
    load it into memory multiple times.
    """

    def _cache_key(self, parser_name: str, idx: str) -> str:
        return f"{self.__class__.__name__}_{parser_name}_{idx}"

    def __init__(self, contexts_path: str, model_path: str):
        """

        :param contexts_path: json file in the format:

            .. code-block:: python

                {"<parser name>": {"<idx>": "<context string>"}}

        :param model_path: path to a pretrained :class:`sklearn.feature_extraction.text.TfidfVectorizer` model
        """
        self.model_path = model_path
        self.contexts_path = contexts_path
        with open(model_path, mode="rb") as f:
            self.vectorizer: TfidfVectorizer = pickle.load(f)
        contexts_path_as_path = Path(contexts_path)
        self._calculate_id_vectors(
            directory=contexts_path_as_path.parent, filename=contexts_path_as_path.name
        )
        self.null_vector = self.vectorizer.transform([""])

    @kazu_disk_cache.memoize(ignore={0, 1})
    def _calculate_id_vectors(self, directory: Path, filename: str) -> None:
        """Calculate the TF-IDF vectors for the file of contexts.

        This method is disk cached - since we don't want every
        single one in memory at runtime, we keep them on disk
        until needed.

        Note that the directory and filename are specified seperately,
        so that disk caching works correctly on different machines.

        :param directory: Directory containing contexts file
        :param filename: contexts json file (see __init__ for format)
        :return:
        """
        with directory.joinpath(filename).open(mode="r") as f:
            data: dict[str, dict[str, str]] = json.load(f)
        for parser_name, ids_and_contexts_dict in data.items():
            mat = self.vectorizer.transform(list(ids_and_contexts_dict.values()))
            with kazu_disk_cache as cache:
                for i, idx in enumerate(ids_and_contexts_dict):
                    cache.set(self._cache_key(parser_name, idx), mat[i])

    @functools.lru_cache(
        maxsize=int(getenv("KAZU_GILDA_TFIDF_DISAMBIGUATION_IN_MEMORY_CACHE_SIZE", 200))
    )
    def _in_memory_disk_cache(self, parser_name: str, idx: str) -> Optional[csr_matrix]:
        return cast(
            Optional[csr_matrix],
            kazu_disk_cache.get(self._cache_key(parser_name=parser_name, idx=idx)),
        )

    def __call__(
        self, context_vec: np.ndarray, id_sets: set[EquivalentIdSet], parser_name: str
    ) -> Iterable[tuple[str, float]]:
        """Given a context vector, yield the most likely identifiers and their score
        from the given set of identifiers.

        :param context_vec:
        :param id_sets:
        :param parser_name:
        :return: identifier strings and scores, starting with the string with the best
            score
        """
        idx_to_vec = {}
        for equiv_id_set in id_sets:
            for idx in equiv_id_set.ids:
                maybe_id_vec = self._in_memory_disk_cache(parser_name=parser_name, idx=idx)
                if maybe_id_vec is not None:
                    idx_to_vec[idx] = maybe_id_vec
                else:
                    idx_to_vec[idx] = self.null_vector
        if idx_to_vec:
            idx_lst = list(idx_to_vec.keys())
            scores = -(
                safe_sparse_dot(
                    context_vec, vstack(idx_to_vec.values()).T, dense_output=True
                ).squeeze()
            )
            neighbours = scores.argsort()
            distances = scores[neighbours]
            for neighbour, score in zip(neighbours, distances):
                idx = idx_lst[neighbour]
                yield idx, -score



import logging
import pickle
import uuid
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass, asdict
from functools import partial
from pathlib import Path
from typing import Any, Union, Optional, cast

import spacy
import srsly
from spacy.language import Language
from spacy.matcher import PhraseMatcher, Matcher
from spacy.tokens import Span, SpanGroup, Doc, Token
from spacy.util import SimpleFrozenList

from kazu.ontology_preprocessing.base import (
    OntologyParser,
)
from kazu.utils.grouping import sort_then_group
from kazu.utils.utils import PathLike

GENE = "gene"
DRUG = "drug"
ANATOMY = "anatomy"
DISEASE = "disease"
CELL_LINE = "cell_line"
CELL_TYPE = "cell_type"

SPAN_KEY = "RAW_HITS"
MATCH_ID_SEP = ":::"

logger = logging.getLogger(__name__)

_CoocDict = dict[str, dict[str, list[str]]]
_MatcherDict = dict[str, Matcher]


@dataclass
class OntologyMatcherConfig:
    span_key: str
    match_id_sep: str
    labels: list[str]
    parser_name_to_entity_type: dict[str, str]


# the strings in the tuple are: parser_name, syn_norm, confidence value
_MatcherOntologyData = dict[str, set[tuple[str, str, str]]]


def _ontology_dict_setter(span: Span, value: dict) -> None:
    # spacy's typing on the property says this has to be a dict[str, Any]
    # but the __init__ typing says it can be a dict[Any, Any]
    span.doc.user_data[span] = value  # type: ignore[index]


def _ontology_dict_getter(span: Span) -> _MatcherOntologyData:
    # as above with types
    return cast(_MatcherOntologyData, span.doc.user_data.get(span, {}))  # type: ignore[call-overload]


@Language.factory(
    "ontology_matcher",
    default_config={
        "span_key": SPAN_KEY,
        "match_id_sep": MATCH_ID_SEP,
    },
)
class OntologyMatcher:
    """String matching to synonyms.

    Core strict matching is done by spaCy's
    `PhraseMatcher <https://spacy.io/api/phrasematcher>`_.
    """

    def __init__(
        self,
        nlp: Language,
        name: str = "ontology_matcher",
        *,
        span_key: str = SPAN_KEY,
        match_id_sep: str = MATCH_ID_SEP,
        parser_name_to_entity_type: dict[str, str],
    ):
        """
        :param nlp: a spacy model, used for its vocab and tokenizer
        :param name: the name of this component. Used for spacy config
        :param span_key: the key for doc.spans to store the matches in
        :param match_id_sep: a separator this splits fields in the match id
        :param parser_name_to_entity_type: a mapping from parsers to their entity class
        """

        Span.set_extension(
            "ontology_dict_", force=True, getter=_ontology_dict_getter, setter=_ontology_dict_setter
        )
        self.nlp = nlp
        self.name = name
        self.cfg = OntologyMatcherConfig(
            span_key=span_key,
            match_id_sep=match_id_sep,
            labels=[],
            parser_name_to_entity_type=parser_name_to_entity_type,
        )
        # These will be defined when calling initialize
        self.strict_matcher: Optional[PhraseMatcher] = None
        self.lowercase_matcher: Optional[PhraseMatcher] = None
        self.tp_matchers: Optional[_MatcherDict] = None
        self.fp_matchers: Optional[_MatcherDict] = None
        self.tp_coocc_dict: Optional[_CoocDict] = None
        self.fp_coocc_dict: Optional[_CoocDict] = None

    @property
    def nr_strict_rules(self) -> int:
        if not self.strict_matcher:
            return 0
        else:
            return len(self.strict_matcher)

    @property
    def nr_lowercase_rules(self) -> int:
        if not self.lowercase_matcher:
            return 0
        else:
            return len(self.lowercase_matcher)

    @property
    def span_key(self) -> str:
        return self.cfg.span_key

    @property
    def match_id_sep(self) -> str:
        return self.cfg.match_id_sep

    @property
    def labels(self) -> list[str]:
        """The labels currently processed by this component."""
        return self.cfg.labels

    @property
    def parser_name_to_entity_type(self) -> dict[str, str]:
        return self.cfg.parser_name_to_entity_type

    def set_labels(self, labels: Iterable[str]) -> None:
        self.cfg.labels = list(labels)
        self.set_context_matchers()

    def set_context_matchers(self):
        self.tp_matchers, self.fp_matchers = self._create_token_matchers()
        self.tp_coocc_dict, self.fp_coocc_dict = self._create_coocc_dicts()

    def create_phrasematchers(
        self, parsers: list[OntologyParser]
    ) -> tuple[Optional[PhraseMatcher], Optional[PhraseMatcher]]:
        """Create spaCy `PhraseMatcher <https://spacy.io/api/phrasematcher>`_\\s.

        :class:`.OntologyStringResource`\\ s are produced by :py:meth:`.OntologyParser.populate_databases`\\ method.

        :param parsers:
        :return:
        """
        if self.strict_matcher is not None or self.lowercase_matcher is not None:
            logging.warning("Phrase matchers are being redefined - is this by intention?")

        self.set_labels(parser.entity_class for parser in parsers)
        strict_matcher = PhraseMatcher(self.nlp.vocab, attr="ORTH")
        lowercase_matcher = PhraseMatcher(self.nlp.vocab, attr="NORM")
        logger.info("ontology matcher build triggered.")
        for parser in parsers:
            parser_resources = parser.populate_databases(return_resources=True)
            if parser_resources is None:
                logger.warning(
                    "tried to create PhraseMatchers from Resources for parser %s, but none have been provided",
                    parser.name,
                )
                continue
            if len(parser_resources) == 0:
                logger.warning(
                    "tried to create PhraseMatchers from Resources for parser %s, but no Resources were produced",
                    parser.name,
                )
                continue

            # deduplicating match id's and patterns saves memory in the spacy pipeline
            match_ids_and_text_cs = set()
            match_ids_and_text_ci = set()
            for resource in parser_resources:
                # a resource can have different syn_norms for different parsers,
                # since the string normalizer's output depends on the entity class.
                # Also, a synonym may exist in multiple LinkingCandidate.raw_synonyms
                syn_norm = resource.syn_norm_for_linking(parser.entity_class)
                for syn in resource.active_ner_synonyms():
                    match_id = (
                        parser.name
                        + self.match_id_sep
                        + syn_norm
                        + self.match_id_sep
                        + str(syn.mention_confidence.value)
                    )
                    if syn.case_sensitive:
                        match_ids_and_text_cs.add(
                            (
                                match_id,
                                syn.text,
                            )
                        )
                    else:
                        match_ids_and_text_ci.add(
                            (
                                match_id,
                                syn.text.lower(),
                            )
                        )

            for match_id, match_str in match_ids_and_text_cs:
                strict_matcher.add(match_id, [self.nlp.tokenizer(match_str)])
            for match_id, match_str in match_ids_and_text_ci:
                lowercase_matcher.add(match_id, [self.nlp.tokenizer(match_str)])

        # only set the phrasematcher if we have any rules for them
        # this lets us skip running a phrasematcher if it has no rules when we come
        # to calling OntologyMatcher
        self.strict_matcher = strict_matcher if len(strict_matcher) != 0 else None
        self.lowercase_matcher = lowercase_matcher if len(lowercase_matcher) != 0 else None
        return self.strict_matcher, self.lowercase_matcher

    def __call__(self, doc: Doc) -> Doc:
        if self.nr_strict_rules == 0 and self.nr_lowercase_rules == 0:
            raise ValueError("there are no matcher rules configured!")

        if self.strict_matcher is not None and self.lowercase_matcher is not None:
            matches = set(self.strict_matcher(doc)).union(set(self.lowercase_matcher(doc)))
        elif self.strict_matcher is not None:
            matches = set(self.strict_matcher(doc))
        elif self.lowercase_matcher is not None:
            matches = set(self.lowercase_matcher(doc))
        else:
            # this isn't possible since if both of them were None,
            # the if clause at the start of this function would raise an error
            raise AssertionError()

        spans: list[Span] = []
        for (start, end), matches_grp in sort_then_group(
            matches,
            key_func=lambda x: (
                x[1],
                x[2],
            ),
        ):
            data = defaultdict(set)
            for mat in matches_grp:
                parser_name, syn_norm, confidence = self.nlp.vocab.strings.as_string(mat[0]).split(
                    self.match_id_sep, maxsplit=2
                )
                ent_class = self.parser_name_to_entity_type[parser_name]
                data[ent_class].add((parser_name, syn_norm, confidence))
            for ent_class, ent_class_data in data.items():
                # we use a uuid here so that every span hash is unique
                new_span = Span(doc, start, end, label=uuid.uuid4().hex)
                new_span._.set("ontology_dict_", {ent_class: ent_class_data})
                spans.append(new_span)
        final_spans = self.filter_by_contexts(doc, spans)
        span_group = SpanGroup(doc, name=self.span_key, spans=final_spans)
        doc.spans[self.span_key] = span_group
        return doc

    def filter_by_contexts(self, doc: Doc, spans: list[Span]) -> list[Span]:
        """These filters work best when there is sentence segmentation available."""
        doc_has_sents = doc.has_annotation("SENT_START")
        # set custom attributes for the token matchers
        for label in self.labels:
            Token.set_extension(label, default=False, force=True)
        # compile the filtered spans by going through each label and each span
        filtered_spans: list[Span] = []
        for label in self.labels:
            # Set all token attributes for other labels (to use in the matcher rules)
            for s in spans:
                for ent_class in s._.ontology_dict_:
                    if ent_class != label and ent_class in self.labels:
                        self._set_token_attr(ent_class=ent_class, span=s, value=True)
            # Set the token attribute for each span of this label one-by-one, then match
            for span in spans:
                for ent_class in span._.ontology_dict_:
                    if ent_class == label:
                        self._set_token_attr(ent_class=ent_class, span=span, value=True)
                        context: Union[Doc, Span] = doc
                        if doc_has_sents:
                            context = span.sent
                        if (
                            self.span_in_TP_coocc(context, span=span, ent_class=ent_class)
                            and not self.span_in_FP_coocc(context, span=span, ent_class=ent_class)
                            and self.span_in_TP_context(context, ent_class)
                            and not self.span_in_FP_context(context, ent_class)
                        ):
                            filtered_spans.append(span)
                        # reset for the next span within the same label
                        self._set_token_attr(ent_class, span, False)
            # reset for the next label
            for s in spans:
                for ent_class in s._.ontology_dict_:
                    if ent_class in self.labels:
                        self._set_token_attr(ent_class, s, False)
        return filtered_spans

    def _set_token_attr(self, ent_class: str, span: Span, value: bool) -> None:
        for token in span:
            token._.set(ent_class, value)

    def span_in_TP_context(self, doc: Union[Doc, Span], ent_class: str) -> bool:
        """When an entity type has a TP matcher defined, it should match for this span
        to be regarded as a true hit."""
        assert self.tp_matchers is not None
        tp_matcher = self.tp_matchers.get(ent_class, None)
        if tp_matcher:
            rule_spans = tp_matcher(doc, as_spans=True)
            if len(rule_spans) > 0:
                return True
            return False
        return True

    def span_in_FP_context(self, doc: Union[Doc, Span], ent_class: str) -> bool:
        """When an entity type has a FP matcher defined, spans that match are regarded
        as FPs."""
        assert self.fp_matchers is not None
        fp_matcher = self.fp_matchers.get(ent_class, None)
        if fp_matcher:
            rule_spans = fp_matcher(doc, as_spans=True)
            if len(rule_spans) > 0:
                return True
            return False
        return False

    def span_in_TP_coocc(self, doc: Union[Doc, Span], span: Span, ent_class: str) -> bool:
        """When an entity type has a TP co-occ dict defined, a hit defined in the dict
        is only regarded as a true hit when it matches at least one of its co-occ
        terms."""
        assert self.tp_coocc_dict is not None
        tp_dict = self.tp_coocc_dict.get(ent_class, None)
        if tp_dict and tp_dict.get(span.text):
            for w in tp_dict[span.text]:
                if w in doc.text:
                    return True
            return False
        return True

    def span_in_FP_coocc(self, doc: Union[Doc, Span], span: Span, ent_class: str) -> bool:
        """When an entity type has a FP co-occ dic defined, a hit defined in the dict is
        regarded as a false positive when it matches at least one of its co-occ
        terms."""
        assert self.fp_coocc_dict is not None
        fp_dict = self.fp_coocc_dict.get(ent_class, None)
        if fp_dict and fp_dict.get(span.text):
            for w in fp_dict[span.text]:
                if w in doc.text:
                    return True
            return False
        return False

    def _create_token_matchers(self) -> tuple[_MatcherDict, _MatcherDict]:
        tp_matchers: _MatcherDict = {}
        fp_matchers: _MatcherDict = {}
        if CELL_LINE in self.labels:
            tp_matchers[CELL_LINE] = self._create_cell_tp_tokenmatcher(CELL_LINE)
        if CELL_TYPE in self.labels:
            tp_matchers[CELL_TYPE] = self._create_cell_tp_tokenmatcher(CELL_TYPE)

        if ANATOMY in self.labels:
            fp_matchers[ANATOMY] = self._create_anatomy_fp_tokenmatcher()
        return tp_matchers, fp_matchers

    def _create_cell_tp_tokenmatcher(self, ent_class: str) -> Matcher:
        """Define patterns where a Cell line or type appears and it's likely a true
        positive."""
        matcher = Matcher(self.nlp.vocab)
        pattern_1: list[dict[str, Any]] = [
            {"_": {ent_class: True}},
            {"LOWER": {"IN": ["cell", "cells"]}},
        ]
        pattern_2: list[dict[str, Any]] = [
            {"LOWER": "cell"},
            {"LOWER": "line"},
            {"_": {ent_class: True}},
        ]
        pattern_3: list[dict[str, Any]] = [
            {"LOWER": "cell"},
            {"LOWER": "type"},
            {"_": {ent_class: True}},
        ]
        matcher.add("Cell_context", [pattern_1, pattern_2, pattern_3])
        return matcher

    def _create_anatomy_fp_tokenmatcher(self) -> Matcher:
        """Define patterns where an atanomy entity appears and it's likely a false
        positive."""
        matcher = Matcher(self.nlp.vocab)
        patterns: list[list[dict[str, Any]]] = []
        if DRUG in self.labels:
            p: list[dict[str, Any]] = [{"_": {DRUG: True}}, {"_": {ANATOMY: True}, "LOWER": "arm"}]
            patterns.append(p)
        p2: list[dict[str, Any]] = [
            {"LOWER": "single"},
            {"LOWER": "-"},
            {"_": {ANATOMY: True}, "LOWER": "arm"},
        ]
        patterns.append(p2)
        p3: list[dict[str, Any]] = [
            {"LOWER": "quality"},
            {"LOWER": "-", "OP": "?"},
            {"LOWER": "of"},
            {"LOWER": "-", "OP": "?"},
            {"_": {ANATOMY: True}, "LOWER": "life"},
        ]
        patterns.append(p3)
        matcher.add("Anatomy_context", patterns)
        return matcher

    def _create_coocc_dicts(self):
        tp_coocc_dict: _CoocDict = {}
        fp_coocc_dict: _CoocDict = {}
        if GENE in self.labels:
            fp_coocc_dict[GENE] = self._create_gene_fp_dict()
        if DISEASE in self.labels:
            fp_coocc_dict[DISEASE] = self._create_disease_fp_dict()
        return tp_coocc_dict, fp_coocc_dict

    _ivf_fertility_treatment_cooccurrence = ["ICSI", "cycle", "treatment"]

    def _create_gene_fp_dict(self):
        """Define cooccurrence links that determine likely FP gene hits."""
        gene_dict = {}
        gene_dict["IVF"] = self._ivf_fertility_treatment_cooccurrence
        return gene_dict

    def _create_disease_fp_dict(self):
        """Define cooccurrence links that determine likely FP disease hits."""
        disease_dict = {}
        disease_dict["MFS"] = ["endpoint"]
        disease_dict["IVF"] = self._ivf_fertility_treatment_cooccurrence
        return disease_dict

    def to_disk(
        self, path: Union[str, Path], *, exclude: Iterable[str] = SimpleFrozenList()
    ) -> None:
        """Serialize the pipe to disk.

        :param path: Path to serialize the pipeline to.
        :param exclude: String names of serialization fields to exclude.
        """

        def pickle_matcher(p, strict):
            with p.open("wb") as outfile:
                if strict:
                    pickle.dump(self.strict_matcher, outfile)
                else:
                    pickle.dump(self.lowercase_matcher, outfile)

        serialize = {}
        serialize["cfg"] = lambda p: srsly.write_json(p, asdict(self.cfg))
        serialize["strict_matcher"] = partial(pickle_matcher, strict=True)
        serialize["lowercase_matcher"] = partial(pickle_matcher, strict=False)
        spacy.util.to_disk(path, serialize, exclude)

    def from_disk(
        self, path: Union[str, Path], *, exclude: Iterable[str] = SimpleFrozenList()
    ) -> "OntologyMatcher":
        """Load the pipe from disk.

        Modifies the object in place and returns it.
        """

        def unpickle_matcher(p, strict):
            with p.open("rb") as infile:
                matcher = pickle.load(infile)
                if strict:
                    self.strict_matcher = matcher
                else:
                    self.lowercase_matcher = matcher

        deserialize = {}

        def deserialize_cfg(path: PathLike) -> None:
            loaded_conf = srsly.read_json(path)
            self.cfg = OntologyMatcherConfig(**loaded_conf)

        deserialize["cfg"] = deserialize_cfg
        deserialize["strict_matcher"] = partial(unpickle_matcher, strict=True)
        deserialize["lowercase_matcher"] = partial(unpickle_matcher, strict=False)
        # spacy's typing says the functions have to only take strings, but they would normally take
        # a path as well without that being a problem.
        spacy.util.from_disk(path, deserialize, exclude)  # type: ignore[arg-type]
        self.set_context_matchers()
        return self



import dataclasses
import logging
from collections import defaultdict, Counter
from typing import Iterable, Optional

try:
    from gliner import GLiNER
    from gliner.modules.token_splitter import WhitespaceTokenSplitter
except ImportError as e:
    raise ImportError(
        "To use GLiNERStep, you need to install gliner.\n"
        "Install with 'pip install kazu[all-steps]'.\n"
    ) from e
from kazu.data import Document, Entity, CharSpan, Section
from kazu.steps import Step, document_iterating_step
from kazu.utils.utils import sort_then_group

logger = logging.getLogger(__name__)


class GLiNERStep(Step):
    """Wrapper for GLiNER models and library. Requires :class:`kazu.data.Section` to
    have sentence spans set on it, as sentences are processed in batches by GLiNER. This
    is to avoid the 'windowing' problem, whereby a multi-token entity could be split
    across two windows, leading to ambiguity over the entity class and spans. Since
    entities cannot theoretically cross sentences, batching sentences eliminates this
    problem.

    If multiple classes are detected for the same string, the most frequently occuring
    one will be selected for all strings (a.k.a majority vote). In the case of a tie,
    the first sequentially detected class will be used.

    .. attention::

       To use this step, you will need `gliner <https://github.com/urchade/GLiNER>`_
       installed, which is not installed as part of the default kazu install
       because this step isn't used as part of the default pipeline.
       You can either do:

       .. code-block:: console

          $ pip install gliner

       Or you can install required dependencies for all steps included in kazu
       with:

       .. code-block:: console

          $ pip install kazu[all-steps]

    Paper:

    | GLiNER: Generalist Model for Named Entity Recognition using Bidirectional Transformer.
    | Urchade Zaratiana, Nadi Tomeh, Pierre Holat, Thierry Charnois
    | https://arxiv.org/abs/2311.08526

    .. raw:: html

       <details>
       <summary>Bibtex Citation Details</summary>

    .. code:: bibtex

       @misc{zaratiana2023gliner,
             title={GLiNER: Generalist Model for Named Entity Recognition using Bidirectional Transformer},
             author={Urchade Zaratiana and Nadi Tomeh and Pierre Holat and Thierry Charnois},
             year={2023},
             eprint={2311.08526},
             archivePrefix={arXiv},
             primaryClass={cs.CL}
       }

    .. raw:: html

       </details>
    """

    def __init__(
        self,
        pretrained_model_name_or_path: str,
        gliner_class_prompt_to_entity_class: dict[str, str],
        threshold: float = 0.3,
    ):
        """

        :param pretrained_model_name_or_path: Passed to ``GLiNER.from_pretrained``. Note that this could attempt to
            download a model from the HuggingFace Hub (see docs for `ModelHubMixin <https://huggingface.co/docs/huggingface_hub/en/package_reference/mixins#huggingface_hub.ModelHubMixin>`_ ).
        :param gliner_class_prompt_to_entity_class: Since GLiNER needs entity class prompts, these might not map exactly
            to our global NER classes. Therefore, this dictionary provides this mapping.
        :param threshold: passed to ``GLiNER.predict_entities``.

        """

        self.model = GLiNER.from_pretrained(pretrained_model_name_or_path)
        self.gliner_class_prompt_to_entity_class = gliner_class_prompt_to_entity_class
        self.threshold = threshold
        self.splitter = WhitespaceTokenSplitter()
        self.max_batch_size = (
            self.model.config.max_len - 10
        )  # -10 to keep a few tokens back for 'special tokens'

    def _create_batches(
        self, section: Section, doc_idx: str
    ) -> Iterable[tuple[CharSpan, CharSpan]]:
        tokens_this_batch = 0
        span_list = list(section.sentence_spans)
        start_span_this_batch: CharSpan = span_list[0]
        end_span_this_batch: Optional[CharSpan] = None
        for sent_span in section.sentence_spans:

            sentence = section.text[sent_span.start : sent_span.end]
            token_count = len(list(self.splitter(sentence)))
            if token_count > self.model.config.max_len:
                logger.warning(
                    "long sentence detected in docid %s. Only the first %s tokens will"
                    "be processed",
                    doc_idx,
                    self.model.config.max_len,
                )
            if (
                tokens_this_batch + token_count >= self.max_batch_size
                and end_span_this_batch is not None
            ):
                yield start_span_this_batch, end_span_this_batch
                tokens_this_batch = 0
                start_span_this_batch = sent_span
            elif sent_span is span_list[-1]:
                yield start_span_this_batch, sent_span

            end_span_this_batch = sent_span
            tokens_this_batch += token_count

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        # needed for majority voting at the end of the document
        section_entities_map = defaultdict(list)
        entity_class_counter: defaultdict[str, Counter[str]] = defaultdict(Counter)
        for section in doc.sections:
            if section.text and not section.sentence_spans:
                logger.warning(
                    "Skipping section of docid %s as %s requires a sentence splitter to have run",
                    doc.idx,
                    self.namespace(),
                )
                continue
            for batch_start_span, batch_end_span in self._create_batches(section, doc.idx):
                predictions = self.model.predict_entities(
                    section.text[batch_start_span.start : batch_end_span.end],
                    labels=self.gliner_class_prompt_to_entity_class.keys(),
                    threshold=self.threshold,
                )
                for ent_pred_dict in predictions:
                    start = ent_pred_dict["start"] + batch_start_span.start
                    end = ent_pred_dict["end"] + batch_start_span.start
                    entity_class = self.gliner_class_prompt_to_entity_class[ent_pred_dict["label"]]
                    match = ent_pred_dict["text"]
                    entity_class_counter[match][entity_class] += 1

                    section_entities_map[section].append(
                        Entity.load_contiguous_entity(
                            start=start,
                            end=end,
                            match=match,
                            entity_class=entity_class,
                            namespace=self.namespace(),
                            metadata={f"{self.namespace()}_score": ent_pred_dict["score"]},
                        )
                    )
        # calculate majority vote and update sections.
        for section, entities in section_entities_map.items():
            for ent_match, ents_this_match in sort_then_group(entities, lambda x: x.match):
                most_common_class_this_match = entity_class_counter[ent_match].most_common(1)[0][0]
                section.entities.extend(
                    dataclasses.replace(ent, entity_class=most_common_class_this_match)
                    for ent in ents_this_match
                )



import functools
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable
from os import getenv
from typing import Optional

import numpy as np
from kazu.data import (
    Document,
    EquivalentIdSet,
    EquivalentIdAggregationStrategy,
    DisambiguationConfidence,
)
from kazu.database.in_memory_db import (
    MetadataDatabase,
    SynonymDatabase,
    NormalisedSynonymStr,
)
from kazu.language.string_similarity_scorers import StringSimilarityScorer
from kazu.ontology_preprocessing.base import DEFAULT_LABEL
from kazu.steps.linking.post_processing.disambiguation.context_scoring import (
    TfIdfScorer,
    GildaTfIdfScorer,
)
from kazu.utils.grouping import sort_then_group
from kazu.utils.string_normalizer import StringNormalizer
from scipy.sparse import csr_matrix

logger = logging.getLogger(__name__)


class DisambiguationStrategy(ABC):
    """The job of a DisambiguationStrategy is to produce a Set of
    :class:`.EquivalentIdSet`\\.

    .. warning::
       The :class:`.EquivalentIdSet`\\s produced needn't map to those contained within
       :attr:`~.LinkingCandidate.associated_id_sets`\\. This may cause confusing behaviour
       during debugging.


    A :meth:`prepare` method is available, which can be cached in the
    event of any duplicated preprocessing work that may be required (see
    :class:`.StrategyRunner` for the complexities of how MappingStrategy
    and DisambiguationStrategy are coordinated).
    """

    def __init__(self, confidence: DisambiguationConfidence):
        """

        :param confidence: the level of confidence that should be assigned to this strategy. This is simply a label
            for human users, and has no bearing on the actual algorithm.
        """
        self.confidence = confidence

    @abstractmethod
    def prepare(self, document: Document) -> None:
        """Perform any preprocessing required.

        :param document:
        :return:
        """
        pass

    @abstractmethod
    def disambiguate(
        self,
        id_sets: set[EquivalentIdSet],
        document: Document,
        parser_name: str,
        ent_match: Optional[str] = None,
        ent_match_norm: Optional[str] = None,
    ) -> set[EquivalentIdSet]:
        """Select a subset of :class:`.EquivalentIdSet`\\.

        :param id_sets: disambiguation result should be based on these id_sets -
            either a standard subset, or subset based on modified :class:`.EquivalentIdSet`
        :param document: source document
        :param parser_name: name of parser that the id_set comes from
        :param ent_match: matched entity string
        :param ent_match_norm: normalised version of entity string
        :return:
        """
        pass

    def __call__(
        self,
        id_sets: set[EquivalentIdSet],
        document: Document,
        parser_name: str,
        ent_match: Optional[str] = None,
        ent_match_norm: Optional[str] = None,
    ) -> set[EquivalentIdSet]:
        self.prepare(document)
        return self.disambiguate(id_sets, document, parser_name, ent_match, ent_match_norm)


class DefinedElsewhereInDocumentDisambiguationStrategy(DisambiguationStrategy):
    """
    1. look for entities on the document that have mappings
    2. filter the incoming set of :class:`.EquivalentIdSet` based on these mappings
    """

    def __init__(self, confidence: DisambiguationConfidence):
        super().__init__(confidence)
        self.mapped_ids: set[tuple[str, str, str]] = set()

    def prepare(self, document: Document) -> None:
        """Note, this method can't be cached, as the state of the document may change
        between executions.

        :param document:
        :return:
        """
        self.mapped_ids = set()
        entities = document.get_entities()
        self.mapped_ids.update(
            (
                mapping.parser_name,
                mapping.source,
                mapping.idx,
            )
            for ent in entities
            for mapping in ent.mappings
        )

    def disambiguate(
        self,
        id_sets: set[EquivalentIdSet],
        document: Document,
        parser_name: str,
        ent_match: Optional[str] = None,
        ent_match_norm: Optional[str] = None,
    ) -> set[EquivalentIdSet]:
        found_id_sets = set()
        for id_set in id_sets:
            filtered_equivalent_id_set_items = set()
            for idx, source in id_set.ids_and_source:
                if (
                    parser_name,
                    source,
                    idx,
                ) in self.mapped_ids:
                    filtered_equivalent_id_set_items.add((idx, source))
            if len(filtered_equivalent_id_set_items) > 0:
                found_id_sets.add(EquivalentIdSet(frozenset(filtered_equivalent_id_set_items)))
        return found_id_sets


class TfIdfDisambiguationStrategy(DisambiguationStrategy):
    """
    1. retrieve all synonyms associated with a :class:`.EquivalentIdSet`, filter out ambiguous ones
       and build a query matrix with the unambiguous ones.
    2. retrieve a list of all detected entity strings from the document, regardless of source and
       build a document representation matrix of these.
    3. perform TFIDF on the query vs document, and sort according to most likely synonym hit from 1.
    4. if the score is above the minimum threshold, create a mapping.

    """

    def __init__(
        self,
        confidence: DisambiguationConfidence,
        scorer: TfIdfScorer,
        context_threshold: float = 0.7,
        relevant_aggregation_strategies: Optional[Iterable[EquivalentIdAggregationStrategy]] = None,
    ):
        """

        :param confidence: the level of confidence that should be assigned to this strategy. This is simply a label
            for human users, and has no bearing on the actual algorithm.
        :param scorer: handles scoring of contexts
        :param context_threshold: only consider synonyms above this search threshold
        :param relevant_aggregation_strategies: Only consider these strategies when selecting synonyms from the
            synonym database, when building a representation. If none, all strategies will be considered
        """
        super().__init__(confidence)
        self.context_threshold = context_threshold
        if relevant_aggregation_strategies is None:
            self.relevant_aggregation_strategies = {EquivalentIdAggregationStrategy.UNAMBIGUOUS}
        else:
            self.relevant_aggregation_strategies = set(relevant_aggregation_strategies)
        self.synonym_db = SynonymDatabase()
        self.scorer = scorer
        self.parser_name_to_doc_representation: dict[str, np.ndarray] = {}

    @functools.lru_cache(maxsize=int(getenv("KAZU_TFIDF_DISAMBIGUATION_DOCUMENT_CACHE_SIZE", 1)))
    def prepare(self, document: Document) -> None:
        """Build document representations by parser names here, and store in a dict.
        This method is cached so we don't need to call it multiple times per document.

        :param document:
        :return:
        """
        parser_names = frozenset(
            candidate.parser_name
            for ent in document.get_entities()
            for candidate in ent.linking_candidates
        )
        self.parser_name_to_doc_representation = self.cacheable_build_document_representation(
            scorer=self.scorer, doc=document, parsers=parser_names
        )

    @staticmethod
    @functools.lru_cache(maxsize=int(getenv("KAZU_TFIDF_DISAMBIGUATION_CACHE_SIZE", 20)))
    def cacheable_build_document_representation(
        scorer: TfIdfScorer, doc: Document, parsers: frozenset[str]
    ) -> dict[str, csr_matrix]:
        """Static cached method, so we don't need to recalculate document representation
        between different instances of this class.

        :param scorer:
        :param doc:
        :param parsers: technically this only has to be a hashable iterable of string - but it should also be
            unique otherwise duplicate work will be done and thrown away, so pragmatically a frozenset makes sense.
        :return:
        """
        strings = " ".join(x.match_norm for x in doc.get_entities())
        res = {}
        for parser in parsers:
            vectorizer = scorer.parser_to_vectorizer[parser]
            res[parser] = vectorizer.transform([strings])
        return res

    def build_id_set_representation(
        self,
        parser_name: str,
        id_sets: set[EquivalentIdSet],
    ) -> dict[NormalisedSynonymStr, set[EquivalentIdSet]]:
        result = defaultdict(set)
        for id_set in id_sets:
            for idx in id_set.ids:

                syns_this_id = self.synonym_db.get_syns_for_id(
                    parser_name,
                    idx,
                    self.relevant_aggregation_strategies,
                )
                for syn in syns_this_id:
                    result[syn].add(id_set)
        return result

    def disambiguate(
        self,
        id_sets: set[EquivalentIdSet],
        document: Document,
        parser_name: str,
        ent_match: Optional[str] = None,
        ent_match_norm: Optional[str] = None,
    ) -> set[EquivalentIdSet]:
        if parser_name not in self.scorer.parser_to_vectorizer:
            return set()

        document_query_matrix = self.parser_name_to_doc_representation[parser_name]
        id_set_representation = self.build_id_set_representation(parser_name, id_sets)
        if len(id_set_representation) == 0:
            return set()
        else:
            indexed_non_ambiguous_syns = list(id_set_representation.keys())
            for best_syn, score in self.scorer(
                strings=indexed_non_ambiguous_syns,
                matrix=document_query_matrix,
                parser=parser_name,
            ):
                if score >= self.context_threshold and len(id_set_representation[best_syn]) == 1:
                    return id_set_representation[best_syn]
            return set()


class GildaTfIdfDisambiguationStrategy(DisambiguationStrategy):
    def __init__(
        self,
        confidence: DisambiguationConfidence,
        scorer: GildaTfIdfScorer,
        context_threshold_delta: float = 0.01,
    ):
        """

        :param confidence:
        :param scorer:
        :param context_threshold_delta: If the maximum delta between the top two :class:`.EquivalentIdSet`\\ s is below this
            value, assume disambiguation has failed
        """
        super().__init__(confidence)
        self.context_threshold_delta = context_threshold_delta
        self.scorer = scorer

    @functools.lru_cache(maxsize=int(getenv("KAZU_TFIDF_DISAMBIGUATION_DOCUMENT_CACHE_SIZE", 1)))
    def prepare(self, document: Document) -> None:
        """Build document representations by parser names here, and store in a dict.

        This method is cached so we don't need to call it multiple times per document.

        :param document:
        :return:
        """
        self.doc_vector = self.cacheable_build_document_representation(
            scorer=self.scorer, doc=document
        )

    @staticmethod
    @functools.lru_cache(maxsize=int(getenv("KAZU_TFIDF_DISAMBIGUATION_CACHE_SIZE", 20)))
    def cacheable_build_document_representation(
        scorer: GildaTfIdfScorer, doc: Document
    ) -> csr_matrix:
        """Static cached method, so we don't need to recalculate document representation
        between different instances of this class.

        :param scorer:
        :param doc:
        :return:
        """
        doc_string = " ".join(x.text for x in doc.sections)
        return scorer.vectorizer.transform([doc_string])

    def disambiguate(
        self,
        id_sets: set[EquivalentIdSet],
        document: Document,
        parser_name: str,
        ent_match: Optional[str] = None,
        ent_match_norm: Optional[str] = None,
    ) -> set[EquivalentIdSet]:

        idx_to_set: defaultdict[str, set[EquivalentIdSet]] = defaultdict(set)
        for id_set in id_sets:
            for idx in id_set.ids:
                idx_to_set[idx].add(id_set)

        best_score = 0.0
        best_set: Optional[set[EquivalentIdSet]] = None
        for idx, score in self.scorer(
            context_vec=self.doc_vector,
            parser_name=parser_name,
            id_sets=id_sets,
        ):
            # Note: we know that len(id_sets)>1 as it's a disambiguation strategy, and
            # it won't be called if len(id_sets)==1
            this_set = idx_to_set[idx]
            if best_set is None:
                best_set = this_set
                best_score = score
            elif best_set is this_set:
                # a lower scoring ID is associated with the same set[EquivalentIDSet]
                continue
            else:
                # we've gotten to the first ID that's actually different
                # to the one with the best score
                if (best_score - score) < self.context_threshold_delta:
                    # delta is insufficient, consider disambiguation to have failed
                    return set()
                else:
                    # the first ID's score is sufficiently above all others
                    return best_set
        # Code should never reach this, but it's required anyway
        return set()


class AnnotationLevelDisambiguationStrategy(DisambiguationStrategy):
    """Certain entities are often mentioned by some colloquial name, even if it's
    technically incorrect.

    In these cases, we may have an annotation_score field in the metadata_db, as a proxy
    of how widely studied the entity is. We use this annotation score as a proxy for
    'given a random mention of the entity, how likely is it that the author is referring
    to instance x vs instance y'. Naturally, this is a pretty unsophisticated
    disambiguation strategy, so should generally only be used as a last resort!
    """

    def prepare(self, document: Document) -> None:
        pass

    def disambiguate(
        self,
        id_sets: set[EquivalentIdSet],
        document: Document,
        parser_name: str,
        ent_match: Optional[str] = None,
        ent_match_norm: Optional[str] = None,
    ) -> set[EquivalentIdSet]:
        best_score = 0
        best_equiv_id_sets = set()

        for id_set in id_sets:
            for idx in id_set.ids:
                score = int(
                    MetadataDatabase().get_by_idx(parser_name, idx).get("annotation_score", 0)
                )
                if score > best_score:
                    best_score = score
                    best_equiv_id_sets = {id_set}
                elif score == best_score:
                    best_equiv_id_sets.add(id_set)

        return best_equiv_id_sets


class PreferDefaultLabelMatchDisambiguationStrategy(DisambiguationStrategy):
    """Prefer ids where the entity match string is the default label (after
    normalisation).

    .. note::
       This strategy is intended to be used with
       :class:`kazu.steps.linking.post_processing.mapping_strategies.strategies.ExactMatchMappingStrategy`
       with the ``disambiguation_essential`` argument
       set to ``True``.
    """

    def __init__(self, confidence: DisambiguationConfidence):
        super().__init__(confidence)
        self.metadata_db = MetadataDatabase()

    def prepare(self, document: Document) -> None:
        pass

    def disambiguate(
        self,
        id_sets: set[EquivalentIdSet],
        document: Document,
        parser_name: str,
        ent_match: Optional[str] = None,
        ent_match_norm: Optional[str] = None,
    ) -> set[EquivalentIdSet]:

        entity_class = self.metadata_db.parser_name_to_ent_class[parser_name]
        disambiguated_id_set = set()
        for equiv_id_set in id_sets:
            for idx, source in equiv_id_set.ids_and_source:
                default_label_norm = StringNormalizer.normalize(
                    self.metadata_db.get_by_idx(idx=idx, name=parser_name)[DEFAULT_LABEL],
                    entity_class=entity_class,
                )
                if default_label_norm == ent_match_norm:
                    disambiguated_id_set.add(
                        (
                            idx,
                            source,
                        )
                    )

        if len(disambiguated_id_set) == 0:
            return set()
        else:
            return {EquivalentIdSet(ids_and_source=frozenset(disambiguated_id_set))}


class PreferNearestEmbeddingToDefaultLabelDisambiguationStrategy(DisambiguationStrategy):
    """Prefer ids where the entity match string is nearest to the default label (as per
    the configured :class:`.StringSimilarityScorer`\\).

    In the case where multiple ID's share the same nearest embedding distance, multiple
    IDs will be returned. This can happen if there are two ids that share the same
    default label.
    """

    def __init__(
        self, complex_string_scorer: StringSimilarityScorer, confidence: DisambiguationConfidence
    ):
        super().__init__(confidence)
        self.complex_string_scorer = complex_string_scorer
        self.metadata_db = MetadataDatabase()

    def prepare(self, document: Document) -> None:
        pass

    def disambiguate(
        self,
        id_sets: set[EquivalentIdSet],
        document: Document,
        parser_name: str,
        ent_match: Optional[str] = None,
        ent_match_norm: Optional[str] = None,
    ) -> set[EquivalentIdSet]:

        assert (
            ent_match is not None
        ), "ent_match is None, but this strategy requires it to function."
        idx_and_scores: set[tuple[tuple[str, str], float]] = set()
        for equiv_id_set in id_sets:

            for idx, source in equiv_id_set.ids_and_source:
                default_label = self.metadata_db.get_by_idx(idx=idx, name=parser_name)[
                    DEFAULT_LABEL
                ]
                score = self.complex_string_scorer(ent_match, default_label)
                idx_and_scores.add(
                    (
                        (
                            idx,
                            source,
                        ),
                        score,
                    )
                )

        result: set[EquivalentIdSet] = set()
        for score, items in sort_then_group(
            idx_and_scores, key_func=lambda x: x[1], reverse=True
        ):  # negative operand for highest score first
            # return on first, as we're only interested in top hit
            result.add(EquivalentIdSet(ids_and_source=frozenset(item[0] for item in items)))
            # we only want the top score
            break

        return result



"""Modified from 'CustomHeaderMiddleware' in the `Starlette Middleware docs
<https://www.starlette.io/middleware/#basehttpmiddleware>`_.

Licensed under BSD 3-Clause along with the rest of Starlette.

Copyright © 2018, `Encode OSS Ltd <https://www.encode.io/>`_.

.. raw:: html

    <details>
    <summary>Full License</summary>

Copyright © 2018, `Encode OSS Ltd <https://www.encode.io/>`_.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

.. raw:: html

    </details>
"""

import logging

from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response

from kazu.web.server import get_request_id

logger = logging.getLogger("ray")


class AddRequestIdMiddleware(BaseHTTPMiddleware):
    """A middleware that puts a request-id from the request's header onto the response.

    This was written to be used in conjunction with
    :class:`.JWTAuthenticationBackend`, which adds a request id header
    to the request as part of authentication.
    """

    def __init__(self, app):
        super().__init__(app)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        req_id = get_request_id(request)
        response = await call_next(request)
        if req_id is not None:
            response.headers["X-request-id"] = req_id

        logger.info("ID: %s Response Code: %s", req_id, response.status_code)
        return response



import dataclasses
import logging
import urllib
from collections.abc import Iterable, Callable
from typing import Protocol, Optional

from kazu.data import (
    Document,
    Entity,
    Mapping,
    StringMatchConfidence,
    DisambiguationConfidence,
    MentionConfidence,
    KazuConfigurationError,
)
from kazu.steps import Step, document_iterating_step
from kazu.utils.grouping import sort_then_group

EntityFilterFn = Callable[[Entity], bool]
MappingFilterFn = Callable[[Mapping], bool]

logger = logging.getLogger(__name__)


class CleanupAction(Protocol):
    def cleanup(self, doc: Document) -> None:
        raise NotImplementedError


class MappingFilterCleanupAction:
    def __init__(self, filter_fns: list[MappingFilterFn]):
        self.filter_fns = filter_fns

    def cleanup(self, doc: Document) -> None:
        for entity in doc.get_entities():
            entity.mappings = {
                mapping
                for mapping in entity.mappings
                if not any(f(mapping) for f in self.filter_fns)
            }


class EntityFilterCleanupAction:
    def __init__(self, filter_fns: list[EntityFilterFn]):
        self.filter_fns = filter_fns

    def cleanup(self, doc: Document) -> None:
        for section in doc.sections:
            section.entities = [
                entity for entity in section.entities if not any(f(entity) for f in self.filter_fns)
            ]


class DropMappingsByConfidenceMappingFilter:
    def __init__(
        self,
        string_match_ranks_to_drop: Iterable[StringMatchConfidence],
        disambiguation_ranks_to_drop: Iterable[DisambiguationConfidence],
    ):
        self.string_match_ranks_to_drop = set(string_match_ranks_to_drop)
        self.disambiguation_ranks_to_drop = set(disambiguation_ranks_to_drop)

    def __call__(self, mapping: Mapping) -> bool:

        return (
            mapping.string_match_confidence in self.string_match_ranks_to_drop
            or mapping.disambiguation_confidence in self.disambiguation_ranks_to_drop
        )


class DropUnmappedEntityFilter:
    def __init__(
        self,
        from_ent_namespaces: Optional[Iterable[str]] = None,
        min_confidence_level: Optional[MentionConfidence] = MentionConfidence.PROBABLE,
    ):
        self.min_confidence_level = min_confidence_level
        self.from_ent_namespaces = None if from_ent_namespaces is None else set(from_ent_namespaces)

    def __call__(self, ent: Entity) -> bool:
        relevant_namespace = (
            self.from_ent_namespaces is None or ent.namespace in self.from_ent_namespaces
        )

        if self.min_confidence_level is None:
            return relevant_namespace and len(ent.mappings) == 0
        else:
            return (
                relevant_namespace
                and len(ent.mappings) == 0
                and ent.mention_confidence < self.min_confidence_level
            )


class LinkingCandidateRemovalCleanupAction:
    def __init__(self):
        pass

    def cleanup(self, doc: Document) -> None:
        for section in doc.sections:
            for ent in section.entities:
                ent.linking_candidates.clear()


class StripMappingURIsAction:
    """Strip the IDs in :class:`kazu.data.Mapping` to just the final part of the URI.

    For example, this will turn
    http://purl.obolibrary.org/obo/MONDO_0004979
    into just MONDO_004979.

    If you don't want URI stripping at all, don't use this Action as part of the CleanupStep/in the pipeline.
    """

    def __init__(self, parsers_to_strip: Optional[Iterable[str]] = None):
        """

        :param parsers_to_strip: if you only want to strip URIs for some parsers and not others,
            provide the parsers to strip here. Otherwise, all parsers will have their IDs stripped.
            This prevents having to keep the full list of parsers in sync here.
        """
        self.parsers_to_strip = parsers_to_strip

    @staticmethod
    def _strip_uri(idx):
        url = urllib.parse.urlparse(idx)
        if url.scheme == "":
            # not a url
            new_idx = idx
        else:
            new_idx = url.path.split("/")[-1]
        return new_idx

    def cleanup(self, doc: Document) -> None:
        for entity in doc.get_entities():
            new_mappings = set()
            for mapping in entity.mappings:
                if (
                    self.parsers_to_strip is not None
                    and mapping.parser_name not in self.parsers_to_strip
                ):
                    new_mappings.add(mapping)
                else:
                    new_mappings.add(dataclasses.replace(mapping, idx=self._strip_uri(mapping.idx)))
            entity.mappings = new_mappings


class DropMappingsByParserNameRankAction(CleanupAction):
    """Removes instances of :class:`.Mapping` based upon some preferential order of
    parsers.

    Useful if you want to filter results based upon some predefined hierarchy of importance,
    for entity classes mapping to multiple parsers. For instance, you may prefer Meddra
    entities over Mondo ones, but will accept Mondo ones if Meddra mappings aren't
    available.

    .. caution::
       To ensure this class is configured correctly, ensure that all
       the parsers you intend to use with it have populated the metadata
       database first. See :meth:`~.OntologyParser.populate_databases`\\.
    """

    def __init__(
        self,
        entity_class_to_parser_name_rank: dict[str, list[str]],
    ):
        """

        :param entity_class_to_parser_name_rank: For a given entity class, only retain the mappings from the first parser
            that an entity has mappings for, based on list ordering (first is preferred).
        """
        self.entity_class_to_parser_name_rank = entity_class_to_parser_name_rank

    def cleanup(self, doc: Document) -> None:
        for entity in doc.get_entities():
            ranks_to_consider = self.entity_class_to_parser_name_rank.get(entity.entity_class)
            if ranks_to_consider is not None:
                try:
                    for _, mappings in sort_then_group(
                        entity.mappings, key_func=lambda x: ranks_to_consider.index(x.parser_name)
                    ):
                        entity.mappings = set(mappings)
                        # only consider the top rank
                        break
                except IndexError:
                    raise KazuConfigurationError(
                        f"Tried to cleanup mappings for {entity}, but at least one mapping has a parser_name that has not been configured with a rank. Mappings: {entity.mappings}"
                    )


class CleanupStep(Step):
    def __init__(self, cleanup_actions: list[CleanupAction]):
        self.cleanup_actions = cleanup_actions

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        for cleanup_action in self.cleanup_actions:
            cleanup_action.cleanup(doc)
